from epstein_files.documents.config.categories.girls import PRUSAKOVA_BERKELY
from epstein_files.documents.config.doc_cfg import DocCfg, EmailCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *

DERSH_GIUFFRE_TWEET = f"about {VIRGINIA_GIUFFRE}"


SOCIAL_CFGS = [
    DocCfg(id='028815', author=INSIGHTS_POD, description=f"business plan", date='2016-08-20', attached_to_email_id='033171'),
    DocCfg(id='011170', author=INSIGHTS_POD, description=f'case study of social media vibe analysis using tweets from #Brexit', date='2016-06-23', attached_to_email_id='033171'),
    DocCfg(id='032324', author=INSIGHTS_POD, description=f"social media election sentiment analysis", date='2016-11-05', attached_to_email_id='032323'),
    DocCfg(id='032281', author=INSIGHTS_POD, description=f"social media election sentiment report", date='2016-10-25', attached_to_email_id='032280'),
    DocCfg(id='028988', author=INSIGHTS_POD, description=f"pitch deck", date='2016-08-20', attached_to_email_id='033171'),
    DocCfg(id='026627', author=INSIGHTS_POD, description=f"report on impact of presidential debate", attached_to_email_id='026626'),
    DocCfg(id='022213', description=f"{SCREENSHOT} Facebook group called 'Shit Pilots Say' disparaging a 'global girl'", is_interesting=False),
    EmailCfg(id='033171', is_interesting=True, comment='Zubair'),
    EmailCfg(
        id='030630',
        author=MARIA_PRUSAKOVA,
        author_reason=PRUSAKOVA_BERKELY,
        description=f'Masha Prusso / {MARIA_PRUSAKOVA} asks about Zubair Khan and {INSIGHTS_POD}',
        is_interesting=True,
    ),
    EmailCfg(
        id='032319',
        description=f"{ZUBAIR_KHAN} says social media vibes are good for Trump in last week before the 2016 election",
        dupe_type='quoted',
        duplicate_ids=['032283'],
        is_interesting=True,
    ),
    EmailCfg(
        id='032325',
        description=f"{ZUBAIR_KHAN} predicts Trump winning in 2016",
        dupe_type='quoted',
        duplicate_ids=['026014'],
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA00719854',
        description=f'{BORIS_NIKOLIC} introduces Epstein to Chris Poole AKA "moot", the founder of 4chan',
        is_interesting=True,
        show_with_name=CHRIS_POOLE
    ),
    EmailCfg(
        id='EFTA00922824',
        description=f"Epstein meets with {CHRIS_POOLE} (founder of 4chan) the day the infamous /pol/ board is created",
        is_interesting=True,
        show_with_name=CHRIS_POOLE,
    ),
]


TWEET_CFGS = [
    DocCfg(id='023050', author=ALAN_DERSHOWITZ, description=DERSH_GIUFFRE_TWEET),
    DocCfg(id='017787', author=ALAN_DERSHOWITZ, description=DERSH_GIUFFRE_TWEET),
    DocCfg(id='033433', author=ALAN_DERSHOWITZ, description=f"{DERSH_GIUFFRE_TWEET} / David Boies", date='2019-03-02'),
    DocCfg(id='033432', author=ALAN_DERSHOWITZ, description=f"{DERSH_GIUFFRE_TWEET} / David Boies", date='2019-05-02'),
    DocCfg(id='031546', author=DONALD_TRUMP, description=f"about Russian collusion", date='2018-01-06'),
    DocCfg(id='030884', author='Ed Krassenstein'),
    DocCfg(id='033236', description=f'selection about Ivanka Trump in Arabic', date='2017-05-20'),
]



from epstein_files.documents.documents.categories import Neutral
from epstein_files.documents.config.doc_cfg import DocCfg
from epstein_files.people.names import THE_REAL_DEAL
from epstein_files.util.constant.strings import (PALM_BEACH, PALM_BEACH_CODE_ENFORCEMENT, PALM_BEACH_TSV,
     PALM_BEACH_PROPERTY_INFO, PALM_BEACH_WATER_COMMITTEE, VIRGIN_ISLANDS)

REAL_DEAL_ARTICLE = 'article by Keith Larsen'


PROPERTY_CFGS = [
    DocCfg(
        id='026759',
        author='Great Bay Condominium Owners Association',
        category=Neutral.PRESSER,
        description=f'Hurricane Irma damage',
        date='2017-09-13',
        is_interesting=False,
    ),
    DocCfg(id='016602', author=PALM_BEACH_CODE_ENFORCEMENT, description='board minutes', date='2008-04-17'),
    DocCfg(id='016554', author=PALM_BEACH_CODE_ENFORCEMENT, description='board minutes', date='2008-07-17', duplicate_ids=['016616', '016574']),
    DocCfg(id='016636', author=PALM_BEACH_WATER_COMMITTEE, description=f"Meeting on January 29, 2009"),
    DocCfg(id='022417', author='Park Partners NYC', description=f"letter to partners in real estate project with architectural plans"),
    DocCfg(id='027068', author=THE_REAL_DEAL, description=f"{REAL_DEAL_ARTICLE} Palm House Hotel Bankruptcy and EB-5 Visa Fraud Allegations"),
    DocCfg(id='029520', author=THE_REAL_DEAL, description=f"{REAL_DEAL_ARTICLE} 'Lost Paradise at the Palm House'", date='2019-06-17'),
    DocCfg(id='016597', author='Trump Properties LLC', description=f'appeal of some decision about Mar-a-Lago by {PALM_BEACH} authorities'),
    DocCfg(id='018743', description=f"Las Vegas property listing"),
    DocCfg(id='016695', description=f"{PALM_BEACH_PROPERTY_INFO} (?)"),
    DocCfg(id='016697', description=f"{PALM_BEACH_PROPERTY_INFO} (?) that mentions Trump"),
    DocCfg(id='016599', description=f"{PALM_BEACH_TSV} consumption (water?)"),
    DocCfg(id='016600', description=f"{PALM_BEACH_TSV} consumption (water?)"),
    DocCfg(id='016601', description=f"{PALM_BEACH_TSV} consumption (water?)"),
    DocCfg(id='016694', description=f"{PALM_BEACH_TSV} consumption (water?)"),
    DocCfg(id='016552', description=f"{PALM_BEACH_TSV} info"),
    DocCfg(id='016698', description=f"{PALM_BEACH_TSV} info (broken?)"),
    DocCfg(id='016696', description=f"{PALM_BEACH_TSV} info (water quality?"),
    DocCfg(
        id='018727',
        description=f"{VIRGIN_ISLANDS} property deal pitch deck, building will be leased to the U.S. govt GSA",
        date='2014-06-01',
    ),
]



"""
The values in the enums match the suffixes of constants.py variables whose names
take the form 'OTHER_FILES_[enum_value]'.
"""
from enum import StrEnum, auto
from itertools import chain
from typing import Type

from epstein_files.util.constant.strings import *

DEFAULT_CATEGORY_STYLE = 'wheat4'
SOCIAL_STYLE = 'khaki3'


class Interesting(StrEnum):
    CRYPTO = auto()
    DIARY = auto()
    GIRLS = auto()
    LETTER = auto()
    MONEY = auto()
    REPUTATION = auto()
    SOCIAL = auto()
    TEXT_MSG = auto()

class Neutral(StrEnum):
    BUSINESS = auto()
    DEPOSITION = auto()
    FINANCE = auto()
    FLIGHT_LOG = auto()
    GOVERNMENT = auto()
    LEGAL = auto()
    MISC = auto()
    PRESSER = auto()
    RESUMÉ = auto()
    SKYPE_LOG = auto()

class Uninteresting(StrEnum):
    ACADEMIA = auto()
    ARTICLE = auto()
    ARTS = auto()
    BOOK = auto()
    CONFERENCE = auto()
    JSON = auto()
    JUNK = auto()
    POLITICS = auto()
    PHONE_BILL = auto()
    PROPERTY = auto()
    TWEET = auto()

# Enum concat from here: https://stackoverflow.com/questions/71527981/combine-two-or-more-enumeration-classes-in-order-to-have-more-levels-access
class Category(StrEnum):
    _ignore_ = 'member cls'
    cls = vars()

    for member in chain(list(Interesting), list(Neutral), list(Uninteresting)):
        cls[member.name] = member.value

    def __str__(self):
        return str(self.value)


# These categories map to highlighted group labels for the purposes of coloring
CATEGORY_STYLE_MAPPING = {
    Interesting.MONEY: Neutral.FINANCE,
    Interesting.REPUTATION: PUBLICIST,
    Interesting.TEXT_MSG: TECH_BRO,
    Neutral.DEPOSITION: VICTIM_LAWYER,
    Neutral.LEGAL: LAWYER,
    Neutral.PRESSER: PUBLICIST,
    Neutral.SKYPE_LOG: TECH_BRO,
    Uninteresting.ARTICLE: JOURNALIST,
    Uninteresting.BOOK: JOURNALIST,
    Uninteresting.CONFERENCE: ACADEMIA,
    Uninteresting.POLITICS: LOBBYIST,
    Uninteresting.PROPERTY: REAL_ESTATE,
}

CATEGORY_STYLES = {
    Interesting.LETTER: 'plum4',
    Interesting.SOCIAL: SOCIAL_STYLE,
    Neutral.FLIGHT_LOG: EPSTEIN_COLOR,
    Neutral.MISC: 'navajo_white1',
    Neutral.RESUMÉ: 'deep_pink4',
    Uninteresting.JSON: 'dark_red',
    Uninteresting.PHONE_BILL: 'slate_blue3',
    Uninteresting.TWEET: SOCIAL_STYLE,
}

# These are the categories we expect to see as OTHER_FILES_[category] variables for in constants.py
CONSTANT_CATEGORIES = [
    c for c in Category if c not in [
        Neutral.BUSINESS,
        Neutral.PRESSER,
        Uninteresting.JSON,
    ]
]

is_interesting = lambda category: _is_in_enum(category, Interesting)
is_neutral = lambda category: _is_in_enum(category, Neutral)
is_uninteresting = lambda category: _is_in_enum(category, Uninteresting)


def is_category(s: str) -> bool:
    return any(fxn(s) for fxn in [is_interesting, is_neutral, is_uninteresting])


def _is_in_enum(category: str, e: Type[StrEnum]) -> bool:
    return bool(next((c for c in e if c == category), False))



"""
Functions that build rich Text links.
"""
from dataclasses import dataclass
from rich.text import Text
from typing import Self

from epstein_files.util.constant.strings import ARCHIVE_LINK_COLOR, ARCHIVE_ALT_LINK_STYLE, ARCHIVE_LINK_COLOR
from epstein_files.util.helpers.rich_helpers import enclose, join_non_empty

EXTERNAL_LINK_STYLE = 'light_slate_grey bold'
OFFICIAL_LINK_STYLE = 'navajo_white3 bold'
SUBSTACK_POST_LINK_STYLE = 'bright_cyan'


@dataclass
class ExternalLink:
    url: str
    comment: str = ''
    comment_style: str = ''
    comment_url: str = ''
    link_style: str = EXTERNAL_LINK_STYLE
    link_text: str = ''
    parentheses_style: str = 'gray27'

    @classmethod
    def official_link(cls, url: str,  **kwargs) -> Self:
        """Alternate constructor."""
        def pop_with_default(arg: str, default_value: str) -> str:
            return kwargs.pop(arg) if arg in kwargs else default_value

        return cls(
            url,
            comment_style=pop_with_default('comment_style', ARCHIVE_ALT_LINK_STYLE),
            link_style=pop_with_default('link_style', OFFICIAL_LINK_STYLE),
            parentheses_style='wheat4',
            **kwargs
        )

    @property
    def link(self) -> Text:
        return link_text_obj(self.url, self.link_text, self.link_style)

    @property
    def link_with_comment(self) -> Text:
        comment = Text('')

        if self.comment_url:
            comment = link_text_obj(self.comment_url, self.comment, self.comment_style)
            comment = enclose(comment, '()')
        elif self.comment:
            comment = enclose(Text(self.comment, 'color(195) dim italic'), '()', self.parentheses_style)

        return join_non_empty(self.link, comment)


def link_markup(
    url: str,
    link_text: str = '',
    style: str | None = ARCHIVE_LINK_COLOR,
    underline: bool = True
) -> str:
    link_text = link_text or url.removeprefix('https://')
    style = ((style or '') + (' underline' if underline else '')).strip()
    return f"[link={url}][{style}]{link_text}[/{style}][/link]"


def link_text_obj(url: str, link_text: str = '', style: str = ARCHIVE_LINK_COLOR) -> Text:
    return Text.from_markup(link_markup(url, link_text, style))


def parenthesize(msg: str | Text, parentheses_style: str = '') -> Text:
    txt = Text(msg) if isinstance(msg, str) else msg
    return Text('(', style=parentheses_style).append(txt).append(')')



#!/usr/bin/env python
"""
Reformat Epstein text message files for readability and count email senders.

    Run: 'EPSTEIN_DOCS_DIR=/path/to/TXT epstein_generate'
"""
import sys
from subprocess import check_output

from dotenv import load_dotenv
load_dotenv()
from rich.markup import escape
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text

from epstein_files.epstein_files import EpsteinFiles, document_cls
from epstein_files.documents.document import Document
from epstein_files.documents.documents.word_count import print_word_counts
from epstein_files.documents.doj_file import DojFile
from epstein_files.documents.email import Email
from epstein_files.documents.messenger_log import MessengerLog
from epstein_files.documents.other_file import OtherFile
from epstein_files.output.epstein_highlighter import highlighter, temp_highlighter
from epstein_files.output.output import (print_curated_chronological, print_doj_files, print_emails_section,
     print_json_files, print_stats, print_other_files_section, print_text_messages_section, print_all_emails_chronological,
     print_email_device_signatures, print_emailers_info, print_json_metadata, show_urls, write_html)
from epstein_files.output.rich import console, print_json, print_subtitle_panel
from epstein_files.output.site.sites import make_clean
from epstein_files.output.title_page import print_color_key, print_title_page_top, print_title_page_bottom
from epstein_files.util.constant.strings import HOUSE_OVERSIGHT_NOV_2025_ID_REGEX
from epstein_files.util.constants import ALL_CONFIGS
from epstein_files.util.env import args, site_config
from epstein_files.util.helpers.data_helpers import flatten, uniquify
from epstein_files.util.helpers.document_helper import diff_files
from epstein_files.util.helpers.file_helper import extract_file_id, is_local_extract_file
from epstein_files.util.logging import exit_with_error, logger
from epstein_files.util.timer import Timer


def epstein_generate() -> None:
    timer, epstein_files = _load_files_and_check_early_exit_args()
    print_title_page_top()

    if args.all_emails_chrono or args.output_word_count or args.output_devices:
        print_color_key()
    else:
        print_title_page_bottom(epstein_files)

    if args.colors_only:
        pass
    elif args.output_devices:
        print_email_device_signatures(epstein_files)
    elif args.output_chrono:
        print_subtitle_panel('Selected Files of Interest in Chronological Order')
        printed_docs = print_curated_chronological(epstein_files)
        timer.log_section_complete('Document', epstein_files.unique_documents, printed_docs)
    elif args.output_word_count:
        print_word_counts(epstein_files)
        timer.print_at_checkpoint(f"Finished counting words")
    elif args.all_doj_files:
        printed_doj_files = print_doj_files(epstein_files)
        timer.log_section_complete('DojFile', epstein_files.doj_files, printed_doj_files)
    else:
        if args.output_emails:
            printed_emails = print_emails_section(epstein_files)
            timer.log_section_complete('Email', epstein_files.emails, printed_emails)
        elif args.all_emails_chrono:
            printed_emails = print_all_emails_chronological(epstein_files)
            timer.log_section_complete('Chronological Email', epstein_files.emails, printed_emails)

        if args.output_texts:
            printed_logs = print_text_messages_section(epstein_files)
            timer.log_section_complete('MessengerLog', epstein_files.imessage_logs, printed_logs)

        if args.output_other:
            printed_files = print_other_files_section(epstein_files)
            timer.log_section_complete('OtherFile', epstein_files.other_files, printed_files)

    write_html(args.build)
    logger.warning(f"Total time: {timer.seconds_since_start_str()}")

    if args.stats:
        print_stats(epstein_files)  # Used for building pytest checks


def epstein_diff():
    """Diff the cleaned up text of two files."""
    diff_files(args.positional_args)


def epstein_grep():
    """Search the cleaned up text of the files."""
    epstein_files = EpsteinFiles.get_files()

    if HOUSE_OVERSIGHT_NOV_2025_ID_REGEX.match(args.positional_args[0]):
        logger.warning(f"'{args.positional_args[0]}' seems to be an ID, running epstein_show instead...")
        epstein_show()
        return

    for search_term in args.positional_args:
        tmp_highlight = temp_highlighter(search_term, 'reverse')
        search_results = epstein_files.docs_matching(search_term, args.names)
        search_results = sorted(search_results, key=lambda sr: sr.document.timestamp_sort_key)
        print_subtitle_panel(f"Found {len(search_results)} documents matching '{search_term}'")
        last_document = None

        for search_result in search_results:
            doc = search_result.document
            lines = search_result.lines

            if (isinstance(doc, Email) and not args.output_emails) \
                    or (isinstance(doc, (DojFile, OtherFile)) and not args.output_other) \
                    or (isinstance(doc, MessengerLog) and not args.output_texts):
                doc.log(f"{type(doc).__name__} Skipping search result...")
                continue
            elif isinstance(doc, Email) and args.email_body:
                lines = [l for l in search_result.lines if l.line_number > doc.header.num_header_rows]

                if not lines:
                    doc.log(f"None of the matches for '{search_term}' seem to be in the body of the email")
                    continue

            if doc.is_duplicate:
                if last_document and not last_document.is_duplicate:
                    console.line()

                last_document = doc
                console.print(doc.duplicate_file_txt)
            elif args.whole_file:
                console.print(doc)
            else:
                console.print(doc._summary_panel)

                for matching_line in lines:
                    line_txt = matching_line.__rich__()
                    console.print(Padding(tmp_highlight(line_txt), site_config.info_padding()), style='gray37')

            console.line()

            if args.debug:
                console.print(doc._debug_txt(), style='dim')
                console.line()

        matched_ids = [result.document.file_id for result in search_results]
        txt = Text("\n\n'").append(search_term, style='bright_red').append("'").append(' matched IDs: ')
        console.print(txt.append(' '.join(matched_ids), style='cyan'))

        if args.repair:
            epstein_files.repair_ids(matched_ids)


def epstein_show():
    """Show the color highlighted file. If --raw arg is passed, show the raw text of the file as well."""
    ids_with_attachments = set([c.attached_to_email_id for c in ALL_CONFIGS])
    raw_docs: list[Document] = []
    console.line()

    try:
        if args.names:
            people = EpsteinFiles.get_files().person_objs(args.names)
            raw_docs = [doc for doc in flatten([p.emails for p in people])]
        else:
            ids = [extract_file_id(arg.upper().strip().strip('_')) for arg in args.positional_args]
            with_attachment_ids = list(set(ids).intersection(ids_with_attachments))
            local_extract_ids = [id for id in ids if is_local_extract_file(id)]
            raw_docs = [Document.from_file_id(id) for id in ids if not is_local_extract_file(id)]

            if local_extract_ids or with_attachment_ids:
                epstein_files = EpsteinFiles.get_files()

                if local_extract_ids:
                    raw_docs += epstein_files.get_ids(local_extract_ids)

                # show the attachments bc reloaded obj won't have them
                for id in with_attachment_ids:
                    existing_doc = epstein_files.get_id(id)
                    logger.warning(f"Showing the attachments now because reloaded Email won't have them:")
                    console.print(existing_doc)

        if any(doc.file_info.has_file for doc in raw_docs):
            # Rebuild the Document objs so we can see result of latest processing
            docs = Document.sort_by_timestamp([document_cls(doc)(doc.file_path) for doc in raw_docs])
        else:
            logger.warning(f"Not reloading derived documents")
            docs = raw_docs

        logger.info(f"Found file IDs {ids} with types: {[doc._class_name for doc in docs]}")
    except FileNotFoundError as e:
        exit_with_error(str(e))
    except Exception as e:
        console.print_exception()
        exit_with_error(str(e))

    for doc in docs:
        if args.open_pdf and doc.file_info.is_doj_file:
            doc.file_info.open_pdf()
        if args.open_txt:
            doc.file_info.open()
        if args.open_url:
            check_output(['open', str(doc.file_info.external_url)])

        console.print('\n', doc)

        if args.raw:
            console.line()
            console.print(Panel(Text(f"RAW {doc.file_id} RAW").append(doc._summary), expand=False, style=doc.border_style))
            console.print(escape(doc.raw_text()), '\n')

            console.print(Panel(Text(f"LINES {doc.file_id} LINES")))
            print(doc._numbered_lines() + '\n\n')

            if isinstance(doc, Email):
                console.print(Panel(Text("actual_text: ").append(doc._summary), expand=False, style=doc.border_style))
                console.print(escape(doc._extract_actual_text()), '\n')

        if args.debug:
            console.print(doc._debug_txt(), style='dim')

    if args.stats:
        print_json("Highlight counts", highlighter.highlight_counts)


def _load_files_and_check_early_exit_args() -> tuple[Timer, EpsteinFiles]:
    if args.make_clean:
        make_clean()
    elif args.show_urls:
        show_urls()
    else:
        timer = Timer()
        epstein_files = EpsteinFiles.get_files(timer)

        if args.emailers_info:
            print_emailers_info(epstein_files)
        elif args.json_metadata:
            print_json_metadata(epstein_files)
        elif args.json_files:
            print_json_files(epstein_files)
        elif args.repair:
            epstein_files.repair_ids(args.positional_args)
        else:
            return timer, epstein_files

    sys.exit()



from epstein_files.documents.documents.categories import Interesting, Neutral
from epstein_files.documents.config.config_builder import letter
from epstein_files.documents.config.doc_cfg import NO_TRUNCATE, CommunicationCfg, DocCfg, EmailCfg
from epstein_files.util.helpers.string_helper import join_truthy, quote
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.logging import logger

PRUSAKOVA_BERKELY = 'Epstein paid for Prusakova to go to Berkeley'
WOMEN_EMPOWERMENT = f"Women Empowerment (WE) conference"


GIRLS_CFGS = [
    DocCfg(id='019300', author=SVETLANA_POZHIDAEVA, description=f'{WOMEN_EMPOWERMENT} f. {KATHRYN_RUEMMLER}', date='2019-04-05', show_full_panel=True),
    DocCfg(id='022267', author=SVETLANA_POZHIDAEVA, description=f'{WOMEN_EMPOWERMENT} founder essay about growing the seminar business'),
    DocCfg(id='022407', author=SVETLANA_POZHIDAEVA, description=f'{WOMEN_EMPOWERMENT} seminar pitch deck'),
    EmailCfg(id='025874', description=f"possibly about {VIRGINIA_GIUFFRE} {QUESTION_MARKS}"),
    EmailCfg(id='030609', duplicate_ids=['030495']),

    # DOJ Files
    letter('EFTA00078198', 'Marsh Law Firm', [USANYS], "allegations of sexual abuse at Epstein's house", '2020-11-10'),
    DocCfg(id='EFTA00306033', author=SERGEY_BELYAKOV, description='Epstein Russian visa', show_full_panel=True),
    DocCfg(id='EFTA00014648', description='former employee "understood that all the girls were school girls"', show_full_panel=True),
    DocCfg(
        id='EFTA00077200',
        author='French Ministry of Justice',
        date='2020-07-08',
        description=f'allegations that {JEAN_LUC_BRUNEL} drugged a model and Epstein + Ghislaine assaulted a minor',
    ),
    DocCfg(
        id='EFTA00079597',
        author='French Ministry of Justice',
        description=f"allegations of sexual assault by {JEAN_LUC_BRUNEL}",
        date='2020-07-08',
    ),
    CommunicationCfg(
        id='EFTA01612733',
        author=MARIA_PRUSAKOVA,
        author_uncertain=True,
        comment='WhatsApp',
        # TODO: highlight_quote
        description=f'says being Epstein\'s lawyer would be "way better than supplying ladies", also discusses Medici Bank',
        show_full_panel=True,
    ),
    CommunicationCfg(
        id='EFTA01616933',
        description='"now im finding pussy for you" / "no one can beat your pussy network"',  # TODO: should be highlight_quote
        show_full_panel=True,
    ),

    EmailCfg(
        id='EFTA02339926',
        author=JULIA_SANTOS,
        author_uncertain='https://www.reddit.com/r/Epstein/comments/1qwbn5i/trafficker_julia_santos/',
        highlight_quote='very sweet and might be naughty t=o',
    ),
    EmailCfg(id='EFTA01781620', author=KIRA_DIKHTYAR, author_uncertain='"Sent from AOL Mobile Mail" in chain'),
    EmailCfg(id='EFTA01805304', author=KIRA_DIKHTYAR, author_reason='Jmail', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01766199', author=KIRA_DIKHTYAR, author_reason='reply', truncate_to=NO_TRUNCATE),
    EmailCfg(
        id='033178',
        author=MARIA_PRUSAKOVA,
        author_reason=PRUSAKOVA_BERKELY,
        description='Masha Prusso asks about Zubair Khan, discusses recruiting girls for Epstein',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA00669538',
        author=MARIA_PRUSAKOVA,
        author_uncertain='discussion of "experiments"',
        description='blowjob instructions, sexual experiments',
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(
        id='EFTA02010589',
        author=MARIA_PRUSAKOVA,
        date='2012-01-31 9:16 PM',
        description=f'raunchy email mentioning "Sasha Grey guy", note {BORIS_NIKOLIC} and bgC3 in SMTP info at end',
        is_interesting=True,
        recipients=[JEFFREY_EPSTEIN],
    ),
    EmailCfg(
        id='EFTA00928879',
        author=MARIA_PRUSAKOVA,
        author_uncertain='sounds like Prusakova',
        description=f'raunchy email, "find fun girls"',
    ),
    EmailCfg(
        id='EFTA01841553',
        author=MARIA_PRUSAKOVA,
        author_uncertain=True,
        highlight_quote='being fucked in the ass hurts me a lot',
        truncate_to=900,
    ),
    EmailCfg(
        id='032374',
        author=MARIA_PRUSAKOVA,
        author_reason=PRUSAKOVA_BERKELY,
        description='boyfriend named Christian mentioned',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA00997627',
        highlight_quote='you will be introduced to more gorgeous Alicas, i promise',
        recipients=[MARIA_PRUSAKOVA],
        recipient_uncertain='Kind regards/Cordialement in signature',
    ),
    EmailCfg(id='032375', author=MARIA_PRUSAKOVA, author_reason=PRUSAKOVA_BERKELY, truncate_to=295),
    EmailCfg(id='033246', author=MARIA_PRUSAKOVA, author_reason='Kind regards/Cordialement in signature'),
    EmailCfg(id='EFTA01990168', author=MARIA_PRUSAKOVA, author_uncertain='"experiments"', is_interesting=True),
    EmailCfg(id='EFTA00671662', author='Miranda', author_reason='quoted signature', description="yet another girl finder"),
    EmailCfg(id='EFTA02441035', author=STEVEN_VICTOR_MD, description="complaints about free medical treatment for Epstein's girls"),
    EmailCfg(
        id='EFTA00659941',
        author=SVETLANA_POZHIDAEVA,
        author_reason='https://tommycarstensen.com/epstein/findings.html#victim-diary-names-multiple-men',
        description=f'{SVETLANA_POZHIDAEVA} forwarding her intimate conversations with {JOSHUA_FINK} to Epstein (to what end?)',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01844498',
        author=SVETLANA_POZHIDAEVA,
        author_reason='mentions Josh (Fink), BlackBerry signature',
        description=f"{JOSHUA_FINK} thinks Svetlana had an affair with Navalny {QUESTION_MARKS}",
    ),
    EmailCfg(
        id='EFTA01868066',
        # author=SVETLANA_POZHIDAEVA,
        # author_uncertain=f"{JOSHUA_FINK} + Sent from Blackberry",
        description=f'impersonating [someone] talking to {JOSHUA_FINK}',
        duplicate_ids=['EFTA01995523'],
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01870453',
        author=SVETLANA_POZHIDAEVA,
        author_uncertain=f"{JOSHUA_FINK} texts",
        description=f'{JOSHUA_FINK} and {SVETLANA_POZHIDAEVA} discuss an abortion ("You have known you are preg for a week")',
    ),
    EmailCfg(id='EFTA01877224', author=SVETLANA_POZHIDAEVA, author_uncertain=f"{JOSHUA_FINK} texts"),
    EmailCfg(id='EFTA00937507', author=SVETLANA_POZHIDAEVA, author_uncertain=f"{JOSHUA_FINK} texts"),
    EmailCfg(
        id='EFTA01991293',
        author=SVETLANA_POZHIDAEVA,
        author_reason=f"convo about {JOSHUA_FINK} breakup",
        description=f'did {ELON_MUSK} date {SVETLANA_POZHIDAEVA} after she blackmailed {JOSHUA_FINK}? ("he asked 6 times if Elon [gave] me anything")',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01998247',
        author=SVETLANA_POZHIDAEVA,
        author_uncertain='device signature',
        description=f'trying to recruit {ELON_MUSK}?',
        is_interesting=True,
        show_with_name=ELON_MUSK,
    ),
    EmailCfg(
        id='EFTA01772677',
        author=SVETLANA_POZHIDAEVA,
        author_uncertain='device signature',
    ),
    EmailCfg(id='EFTA02431535', author=UNKNOWN_GIRL),
    EmailCfg(id='EFTA01875607', author=UNKNOWN_GIRL),
    EmailCfg(id='EFTA00901581', author=UNKNOWN_GIRL),
    EmailCfg(id='EFTA02449477', author=UNKNOWN_GIRL, description='argument about finding girls for Epstein', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01782788', author=UNKNOWN_GIRL, description='"I need 21-24, wiling to travel and work hard"', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02027844', author=UNKNOWN_GIRL, description='permission to "fuck someone with a condom"', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00763537', author=UNKNOWN_GIRL, highlight_quote="19 but I look younger than her"),
    EmailCfg(id='EFTA00738485', recipients=[STEVEN_VICTOR_MD], truncate_to=640),

    # Recipients
    EmailCfg(
        id='EFTA02223525',
        recipients=[ANNA_KASATKINA, BELLA_KLEIN, JANUSZ_BANASIAK],
        recipient_uncertain='https://www.reddit.com/r/Epstein/comments/1qwbn5i/trafficker_julia_santos/',
        truncate_to=800
    ),
    EmailCfg(
        id='EFTA01857628',
        recipients=[RENATA_BOLOTOVA],
        recipient_uncertain='https://x.com/FlippersUpNow/status/2023457206109155345',
        description='"igrushki" means "toys" in Russian',
    ),
    EmailCfg(
        id='EFTA00857669',
        description=f"{SANITA} is maybe the only great girl with right thinking",
        recipients=[SANITA],
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(
        id='EFTA02560884',
        description=f"is Epstein drafting {SVETLANA_POZHIDAEVA}'s responses to {JOSHUA_FINK} {QUESTION_MARKS}",
        recipients=[SVETLANA_POZHIDAEVA],
        recipient_uncertain=f"{JOSHUA_FINK} texts",
    ),
    EmailCfg(
        id='EFTA02543909',
        author_reason='p.selena@yahoo.com in replies',
        description=f'{SVETLANA_POZHIDAEVA}\'s relationship with "J" ({JOSHUA_FINK}?) and "B" {QUESTION_MARKS}',
        recipients=[SVETLANA_POZHIDAEVA],
    ),
    EmailCfg(
        id='EFTA00888467',
        description=f'very strange conversation about "Peter" (maybe {PETER_MANDELSON}?) and "mj" (maybe {HASSAN_JAMEEL}?)',
        is_interesting=True,
        recipients=[UNKNOWN_GIRL],
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(id='EFTA00897668', recipients=[UNKNOWN_GIRL], people=[JEAN_LUC_BRUNEL, JEFFREY_EPSTEIN, UNKNOWN_GIRL]),
    EmailCfg(id='EFTA00848644', recipients=[UNKNOWN_GIRL], is_interesting=True, description='"take a picture of your pussy"'),
    EmailCfg(
        id='EFTA01026268',
        author_reason='juliador89®mail.ru is unredacted',
        description=f"Epstein pressuring {YULIA_DOROKHINA} to find girls in Russia and send him nude pics and videos",
        recipients=[YULIA_DOROKHINA],
        truncate_to=3100,
    ),
    EmailCfg(id='EFTA00078209', description='allegations about Russian organized crime, sex trafficking', truncate_to=8500),
    EmailCfg(id='EFTA01648951', description=f'allegations against Trump, {HOWARD_LUTNICK}, Glen Dubin, {LEON_BLACK}, {JES_STALEY}'),
    EmailCfg(id='EFTA00950368', description='"any [girl] friends for me?"', truncate_to=NO_TRUNCATE),
    EmailCfg(
        id='EFTA00883738',
        description=f'Epstein recruiter {DASHA_GRUPMAN} looking to spend $50,000 to rent a townhouse for "diplomats"',
        show_with_name=DASHA_GRUPMAN,
    ),
    EmailCfg(
        id='EFTA00577409',
        date='2003-06-01',
        date_uncertain='just a guess, string is 4501 which is NULL in Microsoft Outlook',
        description='Epstein and Ghislaine discuss girl "from donad trumps party" who blackmailed [someone]',
    ),
    EmailCfg(
        id='EFTA01877084',
        comment='https://x.com/HansMahncke/status/2024351830998994979',
        description=f'Epstein and {JEAN_LUC_BRUNEL} "managing" {JOSHUA_FINK}\'s relationship with {SVETLANA_POZHIDAEVA}',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA00704924',
        description=f"stalking Josh Fink (attachment is pic of {JOSHUA_FINK} and {FILIPA_PEROVIC} who later married)",
        show_with_name=JOSHUA_FINK,
    ),
    EmailCfg(
        id='EFTA02217672',
        description=f"{DAVID_L_NEUHAUSER} is raising money for a {JOSHUA_FINK} fund, promises Epstein he won't reveal they're talking",
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01907281',
        description=f"Epstein asks about {JOSHUA_FINK}, pretends it's because {TANCREDI_MARCHIOLO} asked *him* (Epstein asked Marchiolo)",
        show_with_name=JOSHUA_FINK,
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA02007083',
        description=f"{BORIS_NIKOLIC} encourages {KIMBAL_MUSK} to ditch his girlfriend for Epstein's girl {JENNIFER_KALIN} (and he says yes)",
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(
        id='EFTA00412260',
        author=LESLEY_GROFF,
        date='2012-05-25 19:51:20',
        description=f'Epstein really wants "*your friend*" (emphasis {LESLEY_GROFF}\'s) to meet {PETER_MANDELSON}',
        truncate_to=4_000,
    ),
    EmailCfg(id='EFTA01047249', author=NADIA_MARCINKO, author_reason='Miro, https://archive.ph/Qa6vU#selection-1621.160-1621.184'),

    # Quotes
    EmailCfg(
        id='EFTA02504370',
        highlight_quote=f'lovely girl with financial troubles',
        recipients=[KIRA_DIKHTYAR],
        recipient_uncertain='"Sent from AOL Mobile Mail"',
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(
        id='EFTA01022353',
        highlight_quote='she said she was 14-15 yo',
        description='(George Models is basically a porn site)',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01885139',
        description=f'({JENNIFER_KALIN}, Epstein\'s girl whom {KIMBAL_MUSK} soon dated)',
        highlight_quote='Kimbal really like Jen',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA02415420',
        highlight_quote="I have a friend of Putin,s",
        is_interesting=True,
    ),
    EmailCfg(id='EFTA01965732', highlight_quote='facilictating his illicit trysts, with married women, to being asked to provide adderall fro bridge tournamnts'),
    EmailCfg(id='EFTA01768670', highlight_quote='find girls for the agency', description=f'but {BORIS_NIKOLIC} is a "biotech investor"...'),
    EmailCfg(id='EFTA01905320', highlight_quote='girls and i are going to see elon musk at space x tomorrow', is_interesting=True),
    EmailCfg(id='EFTA01803353', highlight_quote='having problems with regina', description='(Ramsey?)'),
    EmailCfg(id='EFTA01858707', highlight_quote='How is the marriage plan going'),
    EmailCfg(id='EFTA00950394', highlight_quote='how old will you go?'),
    EmailCfg(id='EFTA02499884', highlight_quote='I gave another girl to kimball and he is thrilled', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02557757', highlight_quote="I have a girl Sana working half time/who by the way you'd like"),
    EmailCfg(id='EFTA00771856', highlight_quote="I know 23 is on the old side for you", is_interesting=True),
    EmailCfg(id='EFTA01974447', highlight_quote='I know you are going to meet putin on the 20th', is_interesting=True),
    EmailCfg(id='EFTA01767237', highlight_quote='is 24 too old for you?', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00854166', highlight_quote='kazak contract will be ready for your review', is_interesting=True, truncate_to=1100),
    EmailCfg(
        id='EFTA00878421',
        highlight_quote="said that she felt gods presence next to her when she was in bed",
        is_interesting=True,
        recipients=[NADIA_MARCINKO],
        recipient_uncertain='https://archive.ph/Qa6vU',  # https://www.nybooks.com/articles/2026/03/26/the-devil-himself-jeffrey-epstein-enright/
    ),
    EmailCfg(id='EFTA00894079', highlight_quote='say hi to Snow White'),
    EmailCfg(id='EFTA00775255', highlight_quote='she is 20 years old but she looks younger', non_participants=[BROCK_PIERCE]),
    EmailCfg(id='EFTA02387676', highlight_quote="thats the closest i get to karyna providing me new pussy"),
    EmailCfg(id='EFTA00908180', highlight_quote='two cinderellas'),
    EmailCfg(id='EFTA01930285', highlight_quote='ukraine upheaval should provide many opportunites'),
    EmailCfg(id='EFTA00751119', highlight_quote='Valdson to teach girls how to serve'),
    EmailCfg(id='EFTA01995972', highlight_quote='you are meeting elon tmr for lunch'),
    EmailCfg(id='EFTA01998027', highlight_quote="What day/night will be the wildest party on your island?", is_interesting=True),
    EmailCfg(id='EFTA01930501', highlight_quote="Your littlest girl was a little naughty"),
    EmailCfg(id='DropSite 2006-05-29 1329', highlight_quote='I am giving the little girl a modeling\n> lesson'),

    # Descriptions
    EmailCfg(id='EFTA01816788', description=f"Russian girl for {PRINCE_ANDREW}"),
    EmailCfg(id='EFTA00572157', description='translation: "Congratulations on the new addition!!! What did you name the little miracle?"'),
    EmailCfg(id='EFTA02559808', description=f"{EVA_DUBIN} delivering amphetamines (adderall) to Epstein", is_interesting=True, truncate_to=200),
    EmailCfg(id='EFTA01840103', description="Epstein apparently suggesting amphetamines (adderall)", is_interesting=True),
    EmailCfg(id='EFTA01987855', description=f'Epstein, Brunel, and {SULTAN_BIN_SULAYEM} very interested in a Liberian sex scandal'),
    EmailCfg(id='EFTA01817903', description=f'Epstein, Brunel, and {SULTAN_BIN_SULAYEM} sharing news of a Liberian sex scandal', is_interesting=True),
    EmailCfg(id='EFTA01818540', description=f'Epstein forwarding news of the Liberian sex scandal to {JES_STALEY}'),
    EmailCfg(id='EFTA01140210', description='Epstein asks about fake Instagram followers', is_interesting=True),
    EmailCfg(id='EFTA00937981', description=f"Epstein looking for {JOSHUA_FINK} (again)"),
    EmailCfg(id='EFTA00668344', description=f"Epstein looking for {JOSHUA_FINK} (again)"),
    EmailCfg(id='EFTA01856467', description=f"Epstein looking for {JOSHUA_FINK} (again)", truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01854125', description=f'Epstein says to {HENRY_JARECKI} "you torture, and mistreat each" girl'),
    EmailCfg(id='EFTA01936438', description=f'Epstein\'s standard password might be "MVEMJSNUP"', is_interesting=True),
    EmailCfg(id='EFTA01953412', description=f'{EVA_DUBIN} suggests Epstein endow "The Epstein Floor For Women" at Mt. Sinai'),
    EmailCfg(id='EFTA00658028', description=f'{FAITH_KATES} trying to get a modeling contract for "Regina"?'),
    EmailCfg(id='EFTA00855566', description=f"introducing {SANITA}", truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA01145923', description=f"{JEAN_LUC_BRUNEL} modeling contract for {MC2_MODEL_MGMT}"),
    EmailCfg(id='EFTA01660651', description='list of Trump accusers', is_interesting=True),
    EmailCfg(id='EFTA02609062', description=f"{MASHA_DROKOVA} is assembling a team", is_interesting=True),
    EmailCfg(id='EFTA02557291', description='possibly recruiting girls'),
    EmailCfg(
        id='EFTA01811230',
        description=f'"Renat" is probably {RENATA_BOLOTOVA}, {REDACTED} might be "Irina" or {SVETLANA_POZHIDAEVA}',
        duplicate_ids=['EFTA00732294'],
    ),
    EmailCfg(id='EFTA00901905', description="smoking in the house oh no!"),
    EmailCfg(id='EFTA00914505', description=f'{SVETLANA_POZHIDAEVA} and {JOSHUA_FINK}...', duplicate_ids=['EFTA01999588']),
    EmailCfg(id='EFTA01798715', description=f"taking an interest in {JOSHUA_FINK}'s investments in Eritrea in between hair gigs"),
    EmailCfg(id='EFTA00631762', description="visa problems for Epstein's South African friend"),
    EmailCfg(
        id='EFTA02434682',
        description=f"Epstein and {JEAN_LUC_BRUNEL} courting a finance bro with women",
        is_interesting=True,
        truncate_to=1500,
    ),
    EmailCfg(
        id='EFTA00927927',
        description=f"reads like {RENATA_BOLOTOVA} talking about recruiting girls for Epstein",
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01745650',
        author=KIRA_DIKHTYAR,
        author_uncertain='AOL signature',
        description=f'{SANITA} is mad at Epstein (and "makes lists")',
        show_with_name=SANITA,
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(id='EFTA00376832', show_with_name=MARIANA_IDZKOWSKA),
    EmailCfg(id='EFTA00743526', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02664956', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00689698', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02702141', truncate_to=NO_TRUNCATE, comment=ELON_MUSK),
    EmailCfg(id='EFTA00819508', truncate_to=NO_TRUNCATE, comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00460312', truncate_to=NO_TRUNCATE, comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00878255', comment=BORIS_NIKOLIC),
    EmailCfg(id='EFTA02551185', comment=BORIS_NIKOLIC),
    EmailCfg(id='EFTA02538269', comment=BORIS_NIKOLIC, duplicate_ids=['EFTA01866636']),
    EmailCfg(id='EFTA00736518', comment=JES_STALEY + ' beauty and the beast'),
    EmailCfg(id='EFTA02344845', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA02491007', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00633284', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00664521', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00997666', comment=JULIA_SANTOS),
    EmailCfg(id='EFTA00964428', comment=JULIA_SANTOS + ' Valdson'),
    EmailCfg(id='EFTA02398135', comment=JULIA_SANTOS + ' "Sochi"'),
    EmailCfg(id='EFTA00766770'),
]



import re
import urllib.parse
from typing import Callable, Literal

from inflection import parameterize
from rich.text import Text

from epstein_files.output.site.sites import GH_PROJECT_URL, TO_FROM, SiteType
from epstein_files.util.env import args
from epstein_files.util.constant.strings import ARCHIVE_ALT_LINK_STYLE, SOCIAL_MEDIA_LINK_STYLE, TEXT_LINK
from epstein_files.util.helpers.file_helper import coerce_file_stem
from epstein_files.util.helpers.link_helper import SUBSTACK_POST_LINK_STYLE, ExternalLink, link_markup, link_text_obj
from epstein_files.util.helpers.string_helper import remove_question_marks

# External site names
ExternalSite = Literal['epstein.media', 'epsteinify', 'EpsteinWeb', 'Jmail', 'RollCall', 'search X']
EPSTEIN_MEDIA = 'epstein.media'
EPSTEIN_WEB = 'EpsteinWeb'
EPSTEINIFY = 'epsteinify'
JMAIL = 'Jmail'
ROLLCALL = 'RollCall'
TWITTER = 'search X'

# External URLs
COFFEEZILLA_ARCHIVE_URL = 'https://journaliststudio.google.com/pinpoint/search?collection=061ce61c9e70bdfd'
COURIER_NEWSROOM_ARCHIVE_URL = 'https://journaliststudio.google.com/pinpoint/search?collection=092314e384a58618'
DOJ_2026_FILE_BASE_URL = "https://www.justice.gov/epstein/files/DataSet%20"
EPSTEIN_DOCS_URL = 'https://epstein-docs.github.io'
OVERSIGHT_REPUBS_PRESSER_URL = 'https://oversight.house.gov/release/oversight-committee-releases-additional-epstein-estate-documents/'
OVERSIGHT_DRIVE_URL = 'https://drive.google.com/drive/folders/1hTNH5woIRio578onLGElkTWofUSWRoH_'
SUBSTACK_URL = 'https://cryptadamus.substack.com/p/i-made-epsteins-text-messages-great'
SUBSTACK_INSIGHTS_POD = 'https://cryptadamus.substack.com/p/maybe-the-russian-bots-were-jeffrey'

SUBSTACK_INSIGHTS_LINK = link_text_obj(
    SUBSTACK_INSIGHTS_POD,
    '"Maybe The Russian Bots Were Jeffrey Epstein The Whole Time"',
    SUBSTACK_POST_LINK_STYLE
)

# DOJ docs
DOJ_2026_URL = 'https://www.justice.gov/epstein/doj-disclosures'
DOJ_SEARCH_URL = 'https://www.justice.gov/epstein/search'
# ours
CRYPTADAMUS_TWITTER = link_markup('https://x.com/cryptadamist', '@cryptadamist')

# Document source sites
EPSTEINIFY_URL = 'https://epsteinify.com'
EPSTEIN_MEDIA_URL = 'https://epstein.media'
EPSTEIN_WEB_URL = 'https://epsteinweb.org'
JMAIL_URL = 'https://jmail.world'

DOC_LINK_BASE_URLS: dict[ExternalSite, str] = {
    EPSTEIN_MEDIA: f"{EPSTEIN_MEDIA_URL}/files/",
    EPSTEIN_WEB: f'{EPSTEIN_WEB_URL}/wp-content/uploads/epstein_evidence/images/',
    EPSTEINIFY: f"{EPSTEINIFY_URL}/document/",
    ROLLCALL: f'https://rollcall.com/factbase/epstein/file?id=',
}

OFFICIAL_LINKS = [
    ExternalLink.official_link(
        DOJ_2026_URL,
        comment='search',
        comment_url=DOJ_SEARCH_URL,
        link_text='DOJ Epstein Files Transparency Act Disclosures'
    ),
    ExternalLink.official_link(
        OVERSIGHT_REPUBS_PRESSER_URL,
        comment='raw files',
        comment_url=OVERSIGHT_DRIVE_URL,
        link_text='2025 Oversight Committee Press Release',
    )
]

EXTERNAL_LINKS = OFFICIAL_LINKS + [
    # ExternalLink(EPSTEIN_WEB_URL, 'biographies', link_text='EpsteinWeb'),
    ExternalLink('https://epsteinexposed.com'),
    ExternalLink(
        'https://randallscott25-star.github.io/epstein-forensic-finance/narratives/19_grand_opus_narrative.html',
        'money',
        link_text='Epstein Audit'
    ),
    ExternalLink(JMAIL_URL, 'read His Emails via Gmail interface', link_text='Jmail'),
    ExternalLink(
        'https://tommycarstensen.com/epstein/index.html',
        comment='findings',
        comment_url='https://tommycarstensen.com/epstein/findings.html',
        comment_style='light_sea_green italic',
        link_text='Carstensen Epstein Archive'
    ),
    ExternalLink(EPSTEIN_DOCS_URL, 'old docs', link_text='epstein-docs'),
    ExternalLink(EPSTEIN_MEDIA_URL, 'raw document images'),
    # ExternalLink(EPSTEINIFY_URL, 'raw images alt', link_text='Epsteinify'),
    ExternalLink('https://bitcoinprotocol.org/epstein-bitcoin-emails', 'crypto', link_text='Epstein Bitcoin Emails'),
    ExternalLink('https://efta-search.vercel.app', 'search filters', link_text='EFTASearch'),
]

CRYPTADAMUS_SOCIAL_LINKS = [
    link_text_obj('https://universeodon.com/@cryptadamist/115572634993386057', '@mastodon', style=SOCIAL_MEDIA_LINK_STYLE),
    link_text_obj(SUBSTACK_URL, '@substack', style=SOCIAL_MEDIA_LINK_STYLE),
    link_text_obj('https://x.com/Cryptadamist/status/1990866804630036988', '@twitter', style=SOCIAL_MEDIA_LINK_STYLE),
    link_text_obj(GH_PROJECT_URL, '@github', style=SOCIAL_MEDIA_LINK_STYLE)
]

# Misc
URL_SIGNIFIERS = ['?amp', 'amp?', 'cd=', 'click', 'CMP=', 'contentId', 'ft=', 'gclid', 'htm', 'mp=', 'keywords=', 'Id=', 'module=', 'mpweb', 'nlid=', 'ref=', 'smid=', 'sp=', 'usg=', 'utm']


# Epsteinify
epsteinify_api_url = lambda file_stem: f"{EPSTEINIFY_URL}/api/documents/{file_stem}"
epsteinify_doc_link_markup = lambda filename_or_id, style = TEXT_LINK: external_doc_link_markup(EPSTEINIFY, filename_or_id, style)
epsteinify_doc_link_txt = lambda filename_or_id, style = TEXT_LINK: Text.from_markup(external_doc_link_markup(filename_or_id, style))
epsteinify_doc_url = lambda file_stem: build_doc_url(DOC_LINK_BASE_URLS[EPSTEINIFY], file_stem)
epsteinify_name_url = lambda name: f"{EPSTEINIFY_URL}/?name={urllib.parse.quote(name)}"

# epstein.media
epstein_media_doc_url = lambda file_stem: build_doc_url(DOC_LINK_BASE_URLS[EPSTEIN_MEDIA], file_stem, 'lower')
epstein_media_doc_link_markup = lambda filename_or_id, style = TEXT_LINK: external_doc_link_markup(EPSTEIN_MEDIA, filename_or_id, style)
epstein_media_doc_link_txt = lambda filename_or_id, style = TEXT_LINK: Text.from_markup(epstein_media_doc_link_markup(filename_or_id, style))
epstein_media_person_url = lambda person: f"{EPSTEIN_MEDIA_URL}/people/{parameterize(person)}"

# EpsteinWeb
epstein_web_doc_url = lambda file_stem: f"{DOC_LINK_BASE_URLS[EPSTEIN_WEB]}/{file_stem}.jpg"
epstein_web_person_url = lambda person: f"{EPSTEIN_WEB_URL}/{parameterize(person)}"
epstein_web_search_url = lambda s: f"{EPSTEIN_WEB_URL}/?ewmfileq={urllib.parse.quote(s)}&ewmfilepp=20"

# Roll Call
rollcall_doc_url = lambda file_stem: build_doc_url(DOC_LINK_BASE_URLS[ROLLCALL], file_stem, 'title')

# Jmail
search_jmail_url = lambda txt: f"{JMAIL_URL}/search?q={urllib.parse.quote(txt)}"
search_twitter_url = lambda txt: f"https://x.com/search?q={urllib.parse.quote(txt)}&src=typed_query&f=live"

PERSON_LINK_BUILDERS: dict[ExternalSite, Callable[[str], str]] = {
    EPSTEIN_MEDIA: epstein_media_person_url,
    EPSTEIN_WEB: epstein_web_person_url,
    EPSTEINIFY: epsteinify_name_url,
    JMAIL: search_jmail_url,
    TWITTER: search_twitter_url,
}


def build_doc_url(base_url: str, filename_or_id: int | str, case: Literal['lower', 'title'] | None = None) -> str:
    file_stem = coerce_file_stem(filename_or_id)
    file_stem = file_stem.lower() if case == 'lower' or EPSTEIN_MEDIA in base_url else file_stem
    file_stem = file_stem.title() if case == 'title' else file_stem
    return f"{base_url}{file_stem}"


def doj_2026_file_url(dataset_id: int, file_stem: str) -> str:
    """justice.gov link e.g. 'https://www.justice.gov/epstein/files/DataSet%208/EFTA00009802.pdf'"""
    return f"{DOJ_2026_FILE_BASE_URL}{dataset_id}/{file_stem}.pdf"


def doj_2026_link_markup(dataset_id, file_stem: str, style: str = TEXT_LINK) -> str:
    url = doj_2026_file_url(dataset_id, file_stem)
    return link_markup(url, file_stem, style)


def jmail_doj_2026_file_url(dataset_id: int, file_stem: str) -> str:
    """Link to Jmail backup of DOJ file."""
    return f"{JMAIL_URL}/drive/vol{dataset_id:05}-{file_stem.lower()}-pdf"


def external_doc_link_markup(site: ExternalSite, filename_or_id: int | str, style: str = TEXT_LINK) -> str:
    url = build_doc_url(DOC_LINK_BASE_URLS[site], filename_or_id)
    return link_markup(url, coerce_file_stem(filename_or_id), style)


def external_doc_link_txt(site: ExternalSite, filename_or_id: int | str, style: str = TEXT_LINK) -> Text:
    return Text.from_markup(external_doc_link_markup(site, filename_or_id, style))


def internal_link_url(search_term: str) -> str:
    """Hack a local link with the `#:~text=` url comment."""
    return f"{this_site_url()}#:~:text={urllib.parse.quote(search_term)}"


def internal_person_link_url(name: str) -> str:
    """Use the 'link to search term' magic URL comment to link to a person's emails, e.g. '#:~:text=to%2Ffrom%20Jack%20Gold"""
    return internal_link_url(f"{TO_FROM} {remove_question_marks(name)}")


def other_site_type() -> SiteType:
    return SiteType.CURATED if args._site_type != SiteType.CURATED else SiteType.EMAILS


def other_site_url() -> str:
    return SiteType.get_url(other_site_type())


def this_site_url() -> str:
    return SiteType.get_url(args._site_type)


THE_OTHER_PAGE_MARKUP = link_markup(other_site_url(), 'the other page', style='light_slate_grey bold')
THE_OTHER_PAGE_TXT = Text.from_markup(THE_OTHER_PAGE_MARKUP)



# Rich reference: https://rich.readthedocs.io/en/latest/reference.html
import json
from copy import deepcopy
from datetime import datetime
from os import devnull
from pathlib import Path
from typing import Mapping, Sequence

from rich import box
from rich.align import Align
from rich.console import Console, Group, RenderableType
from rich.markup import escape
from rich.panel import Panel
from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from rich.theme import Theme

from epstein_files.output.epstein_highlighter import highlighter
from epstein_files.output.highlight_config import HIGHLIGHT_GROUPS
from epstein_files.util.constant.strings import *
from epstein_files.util.constant.urls import *
from epstein_files.util.env import args, site_config
from epstein_files.util.helpers.data_helpers import json_safe, sort_dict
from epstein_files.util.helpers.link_helper import link_markup
from epstein_files.util.helpers.rich_helpers import enclose
from epstein_files.util.logging import logger

NA_TXT = Text(NA, style='dim')
SUBTITLE_PADDING = (1, 0, 1, 0)
GREY_NUMBERS = [58, 39, 39, 35, 30, 27, 23, 23, 19, 19, 15, 15, 15]
VALID_GREYS = [0, 3, 7, 11, 15, 19, 23, 27, 30, 35, 37, 39, 42, 46, 50, 53, 54, 58, 62, 63, 66, 69, 70, 74, 78, 82, 84, 85, 89, 93]

LINK_HREF_LINE_REGEX = re.compile(r"^([>• ]*)(http\S+)(.*)")

DATASET_DESCRIPTION_STYLE = 'gray74'
# INFO_STYLE = 'light_goldenrod2 italic'
INFO_STYLE = 'gray50 italic'
KEY_STYLE = 'dim'
KEY_STYLE_ALT = 'light_steel_blue3'
LAST_TIMESTAMP_STYLE = 'wheat4'
PATH_STYLE = 'deep_pink3'
STR_VAL_STYLE = 'cornsilk1 italic'
STR_VAL_STYLE_ALT = 'light_yellow3 italic'
SYMBOL_STYLE = 'grey70'
TABLE_BORDER_STYLE = 'grey46'
TABLE_TITLE_STYLE = f"gray54 italic"
TITLE_STYLE = 'black on white'  # color(103)'
WARNING_STYLE = 'bold black on white'

DEFAULT_TABLE_KWARGS = {
    'border_style': TABLE_BORDER_STYLE,
    'caption_style': 'navajo_white3 dim italic',
    'header_style': "bold",
    'title_style': TABLE_TITLE_STYLE,
}

THEME_STYLES = {
    DEFAULT: 'wheat4',
    TEXT_LINK: 'deep_sky_blue4 underline',
    # f"{REGEX_STYLE_PREFIX}.{HIGHLIGHTED_QUOTE}": 'on gray27',
    f"{REGEX_STYLE_PREFIX}.{HIGHLIGHTED_QUOTE}": 'honeydew2 italic on gray15',
    f"{REGEX_STYLE_PREFIX}.reverse": 'reverse',
    **{hg.theme_style_name: hg.style for hg in HIGHLIGHT_GROUPS},
}

RAINBOW = [
    'royal_blue1',
    'medium_purple',
    'light_coral',
    'light_slate_gray',
    'dark_goldenrod',
    'wheat4',
    'white',
    'medium_orchid',
    'deep_pink1',
    'navajo_white1',
]

RICH_THEME = Theme(THEME_STYLES)

# Instantiate console object
CONSOLE_KWARGS = {
    'color_system': '256',
    'highlighter': highlighter,
    'record': args.build,
    'safe_box': True,
    'theme': RICH_THEME,
    'width': args.width,
}

if args.suppress_output:
    logger.warning(f"Suppressing terminal output because args.suppress_output={args.suppress_output}...")
    CONSOLE_KWARGS.update({'file': open(devnull, "wt")})

console = Console(**CONSOLE_KWARGS)


def add_cols_to_table(table: Table, cols: list[str | dict], justify: str = 'center') -> None:
    """Left most col will be left justified, rest are center justified."""
    for i, col in enumerate(cols):
        col_justify = 'left' if i == 0 else justify

        if isinstance(col, dict):
            col_name = col['name']
            col_kwargs = deepcopy(col)
            col_kwargs['justify'] = col_kwargs.get('justify', col_justify)
            del col_kwargs['name']
        else:
            col_name = col
            col_kwargs = {'justify': col_justify}

        table.add_column(col_name, **col_kwargs)


def bool_txt(b: bool | None, match_width: bool = False) -> Text:
    txt = Text('')

    if b is False:
        return txt.append(str(b), style='bright_red bold italic')
    elif b is True:
        return txt.append(f" {b}" if match_width else str(b), style='bright_green bold italic')
    else:
        return txt.append(str(b), style='dim italic')


def build_table(
    title: str | Text | None,
    cols: list[str | dict] | None = None,
    copy_props_from: Table | None = None,
    **kwargs
) -> Table:
    table = Table(title=title, **{**DEFAULT_TABLE_KWARGS, **kwargs})

    if copy_props_from:
        for k, v in vars(copy_props_from).items():
            if k.startswith('_') or k in ['columns', 'rows']:
                continue

            setattr(table, k, v)

    if cols:
        add_cols_to_table(table, cols)

    return table


def create_hyperlinks(line: str) -> Text:
    """Add [link] tags if appropriate."""
    if (match := LINK_HREF_LINE_REGEX.match(line)):
        logger.debug(f"Creating hyperlink for '{line}'")
        link = match.group(2)
        txt = Text(match.group(1))
        txt.append(Text.from_markup(f"[link={link}]{link}[/link]"))
        return txt.append(match.group(3))
    else:
        return Text(line)


def indent_txt(txt: str | Text, spaces: int = 4, prefix: str = '') -> Text:
    indent = Text(' ' * spaces).append(prefix)
    return indent + Text(f"\n{indent}").join(txt.split('\n'))


def join_texts(txts: Sequence[str | Text], join: str = ' ', encloser: str = '', encloser_style: str = 'wheat4') -> Text:
    """Join rich.Text objs into one."""
    txt = Text('')

    for i, _txt in enumerate(txts):
        txt.append(join if i >= 1 else '').append(enclose(_txt, encloser, encloser_style))

    return txt


def left_indent(obj: RenderableType, num_spaces: int) -> Padding:
    return Padding(obj, left_indent_padding(num_spaces))


def prefix_with(txt: list[str] | list[Text] | Text | str, pfx: str, pfx_style: str = '', indent: str | int = '') -> Text:
    indent = indent * ' ' if isinstance(indent, int) else indent

    lines = [
        Text('').append(f"{indent}{pfx} ", style=pfx_style).append(line)
        for line in (txt.split('\n') if isinstance(txt, (Text, str)) else txt)
    ]

    return Text('\n').join(lines)


def print_centered(obj: RenderableType, style: str = '') -> None:
    if args.mobile:
        if isinstance(obj, str):
            obj = Text.from_markup(obj, justify='center')
        elif isinstance(obj, Text):
            obj = Text('', justify='center').append(obj)

    console.print(Align.center(obj), highlight=False, style=style)


def print_centered_link(url: str, link_text: str, style: str | None = None) -> None:
    print_centered(link_text_obj(url, link_text, style or ARCHIVE_LINK_COLOR))


def print_json(label: str, obj: object, skip_falsey: bool = False) -> None:
    if isinstance(obj, dict):
        if skip_falsey:
            obj = {k: v for k, v in obj.items() if v}

        obj = json_safe(obj)

    console.line()
    console.print(Panel(label, expand=False))
    console.print_json(json.dumps(obj, sort_keys=True), indent=4)
    console.line()


def print_special_note(note: str | Text) -> None:
    """Print an eye catching panel with extra info about a person."""
    print_centered(Padding(Panel(note, expand=True, padding=(1, 3), style='reverse'), (0, 0, 2, 0)))


def print_subtitle_panel(msg: str, style: str = 'black on white', center: bool = True) -> None:
    """A reverse color panel to put at the top of sections."""
    panel = Panel(
        Text.from_markup(msg, justify='center' if center else 'left'),
        width=site_config.subtitle_width if center else max(len(msg), 150),
        style=style if center else ''
    )

    if center:
        print_centered(Padding(panel, SUBTITLE_PADDING))
    else:
        console.print(panel)


def quote_txt(t: Text | str, try_double_quote_first: bool = False, style: str = '') -> Text:
    if try_double_quote_first:
        quote_char = '"' if '"' not in t else "'"
    else:
        quote_char = "'" if "'" not in t else '"'

    return Text(quote_char, style=style).append(t).append(quote_char)


def styled_dict(
    d: Mapping[str, bool | datetime | str | Path | Text | None],
    key_style: str = KEY_STYLE,
    sep: str = ': ',
    sort_fields: bool = True,
    min_indent: int = 20,
) -> Text:
    """Turn a dict into a colored representation."""
    key_lengths = [len(k) for k in d.keys()] + [min_indent]
    key_prefixes = list(set([k.split('.')[0] for k in d.keys() if '.' in k]))
    key_pfx_colors = {k: RAINBOW[i] for i, k in enumerate(key_prefixes)}

    key_colors = {
        k: key_pfx_colors.get(k.split('.')[0] if '.' in k else k, key_style)
        for k in d.keys()
    }

    return Text('\n').join([
        styled_key_value(k, v, key_style=key_colors.get(k, key_style), indent=max(key_lengths) + 3, sep=sep)
        for k, v in (sort_dict(d) if sort_fields else d.items())
    ])


def styled_key_value(
    key: str,
    val: bool | datetime | int | str | Path | Text | None,
    key_style: str = KEY_STYLE,
    indent: int = 0,
    sep='='
) -> Text:
    """Generate `Text` for 'key=value'."""
    if '.' in key and key_style == KEY_STYLE:
        key_style = KEY_STYLE_ALT

    if val is None:
        val_txt = Text('None', style='dim italic')
    elif isinstance(val, Text):
        val_txt = val
    elif isinstance(val, list):
        val_txt = highlighter(json.dumps(val))
    elif isinstance(val, bool):
        val_txt = bool_txt(val)
    elif isinstance(val, datetime):
        val_txt = Text(str(val), style=TIMESTAMP_STYLE)
    elif isinstance(val, dict):
        val_txt = indent_txt(Text('\n').append(styled_dict(val)), 14)
    else:
        val_txt = None
        val_style = ''

        if isinstance(val, int):
            val = str(val)
            val_style = 'cyan'
        elif isinstance(val, Path):
            val = val = f"'{val}'"
            val_style = PATH_STYLE
        elif isinstance(val, str):
            if val.startswith('http'):
                val_style = ARCHIVE_LINK_UNDERLINE
            elif key.endswith('category') or key == AUTHOR:
                val_txt = highlighter(val)
            elif key.endswith('filename'):
                val_style = PATH_STYLE
            elif key.endswith('style'):
                val_style = f"{val} bold"
            elif key.endswith('_type'):
                val_style = 'light_slate_gray bold'
            elif key.endswith('id'):
                val_style = 'cyan'
            else:
                val_style = STR_VAL_STYLE_ALT if '.' in key else STR_VAL_STYLE
                val_txt = quote_txt(highlighter(val), style=val_style)

        val_txt = val_txt or Text(str(val), style=val_style or 'bright_white')

    txt = Text('').append(f"{key:>{indent}}", style=key_style)
    txt.append(sep, style=SYMBOL_STYLE).append(val_txt)
    return txt


def wrap_in_markup_style(msg: str, style: str | None = None) -> str:
    if style is None or len(style.strip()) == 0:
        return msg

    modifier = ''

    for style_word in style.split():
        if style_word == 'on':
            modifier = style_word
            continue

        style = f"{modifier} {style_word}".strip()
        msg = f"[{style}]{msg}[/{style}]"
        modifier = ''

    return msg


if args.colors_only and args.debug:
    print_json('THEME_STYLES', THEME_STYLES)



"""
Methods that print the title page with all the links etc.
"""
# Rich reference: https://rich.readthedocs.io/en/latest/reference.html
from rich import box
from rich.align import Align
from rich.console import Console, Group, RenderableType
from rich.markup import escape
from rich.panel import Panel
from rich.padding import Padding
from rich.table import Table
from rich.text import Text

from epstein_files.output.epstein_highlighter import highlighter
from epstein_files.output.layout_elements.demi_table import build_demi_table
from epstein_files.output.highlight_config import HIGHLIGHTED_NAMES
from epstein_files.output.rich import *
from epstein_files.output.site.sites import SECTION_ANCHORS
from epstein_files.output.site.site_config import MOBILE_WARNING
from epstein_files.people.names import UNKNOWN
from epstein_files.util.constant.strings import *
from epstein_files.util.constant.urls import *
from epstein_files.util.constants import HEADER_ABBREVIATIONS
from epstein_files.util.env import args, site_config
from epstein_files.util.helpers.link_helper import SUBSTACK_POST_LINK_STYLE, link_markup, link_text_obj, parenthesize
from epstein_files.util.helpers.rich_helpers import vertically_pad
from epstein_files.util.helpers.string_helper import starred_header
from epstein_files.util.logging import logger

DATASET_DESCRIPTION_STYLE = 'gray74'
DIRECTORY_STYLE = 'wheat4'
OTHER_PAGE_MSG_STYLE = 'gray78 dim'
STR_VAL_STYLE = 'cornsilk1 italic'
STR_VAL_STYLE_ALT = 'light_yellow3 italic'
SECTION_HEADER_STYLE = 'bold black on color(146)'

SITE_GLOSSARY_MSG = f"The following views of the underlying selection of Epstein Files are available:"
YOU_ARE_HERE = Text('«').append('you are here', style='bold khaki1 blink').append('»')
SECTION_LINK_MSG = 'jump to a different section of this page'
SUBTITLE_WIDTH = 110
TITLE_WIDTH = 50

# label of the HighlightedNames objects colors with the style of that same HighlightedNames
COLOR_KEYS = [
    Text(highlight_group.label.replace('_', ' '), style=highlight_group.style)
    for highlight_group in sorted(HIGHLIGHTED_NAMES, key=lambda hg: hg.label)
]

SECTION_LINKS = [
    link_text_obj(internal_link_url(anchor), section_name, 'indian_red')
    for section_name, anchor in SECTION_ANCHORS.items()
]


def print_color_key() -> None:
    color_table = build_table('Rough Guide to Highlighted Colors', show_header=False)
    num_colors = len(COLOR_KEYS)
    row_number = 0

    for i in range(0, site_config.num_color_key_cols):
        color_table.add_column(f"color_col_{i}", justify='center')

    while (row_number * site_config.num_color_key_cols) < num_colors:
        idx = row_number * site_config.num_color_key_cols
        color_table.add_row(*COLOR_KEYS[idx:(idx + site_config.num_color_key_cols)])
        row_number += 1

    print_centered(vertically_pad(color_table))


def print_other_page_link(epstein_files: 'EpsteinFiles') -> None:
    if args._site_type == SiteType.CURATED_MOBILE:
        return
    elif args._site_type == SiteType.CURATED:
        txt = THE_OTHER_PAGE_TXT + Text(f' is uncurated and has all {len(epstein_files.emails):,} emails')
        txt.append(f" and {len(epstein_files.other_files):,} unclassifiable files")
    else:
        txt = THE_OTHER_PAGE_TXT + (f' displays a curated collection of emails and')
        txt.append(" unclassifiable files of particular interest")

    print_centered(parenthesize(txt), style=OTHER_PAGE_MSG_STYLE)
    chrono_emails_link = link_text_obj(SiteType.get_url(SiteType.EMAILS_CHRONOLOGICAL), 'a page', style='light_slate_grey bold')
    chrono_emails_txt = Text(f"there's also ").append(chrono_emails_link)
    chrono_emails_txt.append(' with all the emails in chronological order')
    print_centered(parenthesize(chrono_emails_txt), style=OTHER_PAGE_MSG_STYLE)


def print_section_header(msg: str, style: str = SECTION_HEADER_STYLE, is_centered: bool = False) -> None:
    if args._site_type == SiteType.CURATED:
        console.line(2)
        print_section_links()

    panel = Panel(Text(msg, justify='center'), expand=True, padding=(1, 1), style=style)
    panel = Align.center(panel) if is_centered else panel
    console.print(Padding(panel, site_config.section_header_padding))


def print_section_links(style: str = '') -> None:
    """Print links to the various sections within the curated page."""
    print_centered(build_demi_table(SECTION_LINK_MSG, SECTION_LINKS), style=style)


def print_title_page_top() -> None:
    """Top half of the title page."""
    _print_page_title()
    print_centered(_site_directory())
    console.line()
    _print_starred_header(site_config.not_all_files_warning, num_spaces=0, num_stars=0)
    print_centered(f"This dataset includes everything from the {HOUSE_OVERSIGHT_TRANCHE}", style=DATASET_DESCRIPTION_STYLE)
    print_centered(f"as well as a curated selection of the {DOJ_2026_TRANCHE}", style=DATASET_DESCRIPTION_STYLE)
    print_centered(f"with a particular focus on cryptocurrency", style=DATASET_DESCRIPTION_STYLE)
    print_centered(f"and women who speak Russian.", style=DATASET_DESCRIPTION_STYLE)


def print_title_page_bottom(epstein_files: 'EpsteinFiles') -> None:
    """Bottom half of the title page."""
    console.line()
    _print_abbreviations_table()
    _print_external_links()
    console.line(2)
    print_centered(epstein_files.overview_table())
    # updated_txt = Text('last updated: ').append(highlighter(datetime.now().date().isoformat()))
    # updated_panel = Panel(updated_txt, expand=False, style='grey35', padding=(0, 4))
    # console.line()
    # print_centered(updated_panel)
    print_color_key()
    print_centered(f"(if you think there's an attribution error or can deanonymize an {UNKNOWN} contact {CRYPTADAMUS_TWITTER})", 'grey46')
    print_centered(f"(thanks to {link_markup('https://x.com/ImDrinknWyn', '@ImDrinknWyn', 'dodger_blue3')} + others for help attributing redacted emails)")
    console.line()


def _bulleted_site_link(site_type: SiteType, link: Text) -> Text:
    """These go in the site directory panel."""
    you_are_here = ''

    if args.mobile:
        if site_type == SiteType.CHRONOLOGICAL_MOBILE and args.output_chrono:
            you_are_here = YOU_ARE_HERE
        elif site_type == SiteType.CURATED:
            you_are_here = YOU_ARE_HERE

        if site_type not in SiteType.mobile_compatible_types():
            link.stylize('dim strike')
    elif site_type == args._site_type:
        you_are_here = YOU_ARE_HERE

    return Text('➱ ').append(link).append(' ').append(you_are_here)


def _print_abbreviations_table() -> None:
    if args.mobile:
        return

    table = build_table(title="Abbreviations Used Frequently In These Conversations", show_header=False)
    table.add_column("Abbreviation", justify="center", style='bold')

    table.add_column(
        "Translation",
        justify="center",
        min_width=site_config.abbreviations_width,
        style="white"
    )

    for k, v in HEADER_ABBREVIATIONS.items():
        table.add_row(highlighter(k), v)

    console.print(Align.center(vertically_pad(table)))


def _print_external_links() -> None:
    print_centered(Text('External Links', style=TABLE_TITLE_STYLE))

    for link in sorted(EXTERNAL_LINKS, key=lambda link: -len(link.link_with_comment)):
        print_centered(link.link_with_comment)


def _print_page_title(width: int = TITLE_WIDTH) -> None:
    print_centered(MOBILE_WARNING, style='dim')
    title = Text('', justify='center').append('The Epstein Files', style='underline bold')
    title_panel = Panel(title, box=box.DOUBLE_EDGE, expand=True, padding=(2, 2), style=TITLE_STYLE, width=width)
    print_centered(vertically_pad(title_panel))

    print_centered_link(
        SUBSTACK_URL,
        "I Made Epstein's Text Messages Great Again (And You Should Read Them)",
        style=f'{SUBSTACK_POST_LINK_STYLE} bold'
    )

    if not args.mobile:
        print_centered_link(SUBSTACK_URL, SUBSTACK_URL.removeprefix('https://'), style=f'{SUBSTACK_POST_LINK_STYLE} dim')

    print_centered(join_texts(CRYPTADAMUS_SOCIAL_LINKS, join=site_config.social_link_separator))
    console.line()


def _print_starred_header(msg: str, num_stars: int = 7, num_spaces: int = 2, style: str = WARNING_STYLE) -> None:
    """String like '  *** Title Msg ***  '."""
    print_centered(wrap_in_markup_style(starred_header(msg, num_stars, num_spaces), style))


def _site_directory() -> Panel:
    """Panel with links to the various sites generated by this code."""
    links_txts = [
        _bulleted_site_link(site_type, link)
        for site_type, link in SiteType.directory_links().items()
    ]

    if args.mobile:
        num_link_indent_spaces = 0
    else:
        max_link_len = max(len(link.plain) for link in links_txts)
        num_link_indent_spaces = max(2, int((len(SITE_GLOSSARY_MSG) - max_link_len) / 2)) - 2

    sites_txt = Text('').append(SITE_GLOSSARY_MSG, style=f'{DIRECTORY_STYLE} bold').append('\n\n')
    sites_txt.append(indent_txt(join_texts(links_txts, '\n'), num_link_indent_spaces))

    if args.mobile:
        sites_txt.append("\n\nNot all views are available on mobile.", DIRECTORY_STYLE)

    return Panel(
        sites_txt,
        border_style='dim',
        padding=(1, site_config.site_glossary_horizontal_padding)
    )



from dataclasses import dataclass, field
from typing import Literal

from rich.align import Align
from rich.console import Console, ConsoleOptions, RenderResult, RenderableType
from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from rich.panel import Panel

from epstein_files.output.html.elements import BLACK_BACKGROUND, HtmlStyle, div_class, div_with_legend
from epstein_files.output.html.builder import (PANEL_BASE_PROPS, VERTICAL_MARGIN, border_css_props, rich_to_html,
     one_row_table_html, text_to_list, text_to_div, vertical_margin_props)
from epstein_files.output.html.elements import div_tag, to_em, side_props
from epstein_files.output.rich import join_texts

JustifyMethod = Literal['center', 'left', 'right']

BOTTOM_PADDING = 1
SUBHEADER_VERTICAL_PADDING = to_em(0.3)

DOC_DIV_CSS_PROPS = {
    'display': 'flex',
    'flex-direction': 'column',
}


@dataclass(kw_only=True)
class BasePanel:
    border_style: str
    text: Text | list[Text]
    title: Text | None = None
    title_justify: JustifyMethod = 'right'

    @property
    def is_list(self) -> bool:
        return isinstance(self.text, list)

    def to_div(self, indents: tuple[int, int] | None = None) -> str:
        """Create an HTML <div> string for this panel."""
        indents = indents or (0, 0)
        div_props = {}

        if indents[0]:
            div_props.update(side_props('margin', ['left'], to_em(indents[0])))

        if indents[1]:
            div_props.update(side_props('margin', ['right'], to_em(indents[1])))

        # Handle MessengerLog / TextMessage list. # TODO this sucks
        if self.is_list:
            html = text_to_list(self.text, class_name='no_bullets')
            div_props.update({'word-wrap': 'break-word'})
            return div_class(html, BLACK_BACKGROUND, div_props)

        html = rich_to_html(self.text)
        div_props.update(PANEL_BASE_PROPS)
        div_props.update(border_css_props(self.border_style))

        # TODO: make the title 'dim'
        title = self.title.plain if self.title else ''
        return div_with_legend(html, title, div_props, class_name=BLACK_BACKGROUND)

    def __rich__(self) -> Panel:
        return Panel(
            join_texts(self.text, '\n') if self.is_list else self.text,
            border_style=self.border_style,
            expand=False,
            title=self.title,
            title_align=self.title_justify,
        )


@dataclass(kw_only=True)
class FileDisplay:
    """Allows for proper right vs. left justify of a Document display."""
    background_color: str = ''
    body_panel: BasePanel | Table
    file_info: BasePanel | None = None
    indent: int = 0  # TODO: this should be a property of the BasePanel
    justify: JustifyMethod | None = None
    margin_bottom: str = VERTICAL_MARGIN
    subheaders: list[Text] = field(default_factory=list)

    @property
    def horizontal_body_margin(self) -> tuple[int, int]:
        """(left, right)"""
        return (self.margin[3], self.margin[1])

    @property
    def horizontal_body_margin_css_props(self) -> dict[str, str]:
        """(left, right)"""
        props = {}

        if self.horizontal_body_margin[0]:
            props.update(side_props('margin', ['left'], to_em(self.horizontal_body_margin[0])))
        if self.horizontal_body_margin[1]:
            props.update(side_props('margin', ['right'], to_em(self.horizontal_body_margin[1])))

        return props

    @property
    def is_table(self) -> bool:
        return isinstance(self.body_panel, Table)

    @property
    def justified_subheaders(self) -> list[Text]:
        """Set justify `Text` to get right justification of text in right aligned elements."""
        subheaders = [txt.copy() for txt in self.subheaders]

        if self.justify:
            for txt in subheaders:
                txt.justify = self.justify

        return subheaders

    # TODO: rename body_margin?
    @property
    def margin(self) -> tuple[int, int, int, int]:
        """Margin for the body."""
        padding = [0, 0, 0, 0]

        # Set subtle indent
        if self.justify == 'right':
            padding[1] = self.indent
        else:
            padding[3] = self.indent

        return tuple(padding)

    @property
    def subheader_div(self) -> str:
        if not self.subheaders:
            return ''

        css_props = {
            **self.horizontal_body_margin_css_props,
            **vertical_margin_props(SUBHEADER_VERTICAL_PADDING),
        }

        return text_to_div(Text('\n').join(self.subheaders), css_props)

    def align(self, element: RenderableType) -> RenderableType:
        return Align(element, self.justify) if self.justify else element

    def to_html(self) -> str:
        if self.is_table:
            body_html = one_row_table_html(self.body_panel, self.horizontal_body_margin_css_props)
        else:
            body_html = self.body_panel.to_div(self.horizontal_body_margin)

        elements = [self.file_info.to_div()] if self.file_info else []
        elements.extend([self.subheader_div, body_html])

        inner_html = '\n'.join(elements)
        div_props = {**DOC_DIV_CSS_PROPS, 'margin-bottom': self.margin_bottom}

        if self.background_color:
            div_props.update(HtmlStyle(f"on {self.background_color}").to_css)

        # Add more vertical margin before/after text messages
        if isinstance(self.body_panel, BasePanel) and self.body_panel.is_list:
            div_props.update(vertical_margin_props(to_em(2)))

        return div_tag(inner_html, div_props)

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        """Default `Document` renderer (Email and MessengerLog override this)."""
        # Set justify on the Text in the body panel
        if self.justify and not self.is_table and isinstance(self.body_panel.text, Text):
            self.body_panel.text.justify = self.justify

        indented_elemeents = [*self.justified_subheaders, self.body_panel]
        indented_elemeents = [Padding(e, self.margin) for e in indented_elemeents]
        indented_elemeents[-1].bottom = BOTTOM_PADDING
        elements = ([self.file_info] if self.file_info else []) + indented_elemeents

        for element in elements:
            yield self.align(element)



from rich.text import Text

from epstein_files.people.names import *
from epstein_files.util.constant.strings import PALM_BEACH_CODE_ENFORCEMENT, PALM_BEACH_WATER_COMMITTEE, UN_GENERAL_ASSEMBLY
from epstein_files.util.constant.urls import SUBSTACK_INSIGHTS_LINK

# People who are of interest as authors of non-emails
AUTHORS_OF_INTEREST = [
    DOJ,
    EDWARD_JAY_EPSTEIN,
    FBI,
    INSIGHTS_POD,
    # NOAM_CHOMSKY,
    MICHAEL_WOLFF,
]

# List of people whose files to print in curated site. Order matters.
EMAILERS_OF_INTEREST = [
    HOWARD_LUTNICK,
    BROCK_PIERCE,
    HAMAD_BIN_JASSIM,
    JOI_ITO,
    STEVEN_SINOFSKY,
    AMIR_TAAKI,
    DANIEL_SIAD,
    AL_SECKEL,
    AUSTIN_HILL,
    VINCENZO_IOZZO,
    MARIA_PRUSAKOVA,
    JEREMY_RUBIN,
    JEAN_LUC_BRUNEL,
    DAVID_STERN,
    STEVEN_HOFFENBERG,
    ALAN_DLUGASH,
    EHUD_BARAK,
    MADARS_VIRZA,
    JASON_CALACANIS,
    MASHA_DROKOVA,
    STEVE_BANNON,
    JULIA_SANTOS,
    TYLER_SHEARS,
    SERGEY_BELYAKOV,
    GANBAT_CHULUUNKHUU,
    RENATA_BOLOTOVA,
    CHRISTINA_GALBRAITH,
    PHILIP_ROSEDALE,
    GINO_YU,
    ANDREW_FARKAS,
    ARIANNA_SIMPSON,
    SVETLANA_POZHIDAEVA,
    MOHAMED_WAHEED_HASSAN,
    JENNIFER_JACQUET,
    ZUBAIR_KHAN,
    UMAR_DZHABRAILOV,
    FAWZI_SIAM,
    ROSS_GOW,
    DAVID_BLAINE,
    KARIM_WADE,
    PETER_THIEL,
    UNKNOWN_GIRL,
    ELON_MUSK,
    YULIA_DOROKHINA,
    CHRIS_POOLE,
    RASSECK_BOURGI,
    IAN_OSBORNE,
    DASHA_GRUPMAN,
    BOBBY_KOTICK,
    KIRA_DIKHTYAR,
    JOSHUA_FINK,
    SANITA,
    None,
]

EMAILERS_OF_INTEREST_SET: set[Name] = set(EMAILERS_OF_INTEREST)
EMAILERS_TO_PRINT = EMAILERS_OF_INTEREST + [JEFFREY_EPSTEIN]

TEXTERS_OF_INTEREST = [
    ANTHONY_SCARAMUCCI,
    ANIL_AMBANI,
    STACEY_PLASKETT,
]

PERSONS_OF_INTEREST: set[Name] = set(EMAILERS_OF_INTEREST + AUTHORS_OF_INTEREST + TEXTERS_OF_INTEREST)

UNINTERESTING_AUTHORS = [
    GORDON_GETTY,
    LAWRENCE_KRAUSS,
    NOBEL_CHARITABLE_TRUST,
    PALM_BEACH_CODE_ENFORCEMENT,
    PALM_BEACH_WATER_COMMITTEE,
    UN_GENERAL_ASSEMBLY,
]

SPECIAL_NOTES = {
    ZUBAIR_KHAN: Text('', 'bold', justify='center').append(SUBSTACK_INSIGHTS_LINK) + \
        Text('\n(a post about the social media work Mr. Khan was doing for Epstein during the 2016 election)', 'italic'),
}



from dataclasses import dataclass, field
from email import policy
from email.message import EmailMessage
from email.parser import BytesParser

from epstein_files.documents.emails.email_header import EmailHeader
from epstein_files.documents.email import Email
from epstein_files.util.logging import logger
from epstein_files.util.constant.strings import AUTHOR
from epstein_files.util.helpers.string_helper import collapse_newlines

HEADERS_TO_CHECK = ['Authentication-Results', 'Content-Transfer-Encoding', 'Content-Type', 'DKIM-Signature', 'DomainKey-Signature', 'Received', 'References', 'X-Mailer']
DEFAULT_EML_HEADERS = HEADERS_TO_CHECK + ['Date', 'From', 'Subject', 'To', 'MIME-Version', 'Content-Length']
HEADER_FIELDS = [AUTHOR, 'sent_at', 'to', 'subject']


@dataclass
class DropsiteEmail(Email):
    _eml: EmailMessage = field(init=False)

    @property
    def eml(self) -> EmailMessage:
        if '_eml' not in dir(self):
            with open(self.file_path, 'rb') as fp:
                self._eml = BytesParser(policy=policy.default).parse(fp)

        return self._eml

    def _extract_header(self) -> EmailHeader:
        """Extract an `EmailHeader` from the OCR text."""
        bcc = self.eml['bcc'].split(';') if self.eml['bcc'] else None
        cc = self.eml['cc'].split(';') if self.eml['cc'] else None
        field_names = HEADER_FIELDS + (['bcc'] if bcc else []) + (['cc'] if cc else [])

        for header in self.eml.keys():
            if header not in DEFAULT_EML_HEADERS:
                self.warn(f"unexpected header {header}: '{self.eml[header]}'")

        return EmailHeader(
            author=self.eml['from'],
            bcc=bcc,
            cc=cc,
            field_names=field_names,
            sent_at=self.eml['date'],
            subject=self.eml['subject'],
            to=self.eml['to'].split(';') if self.eml['to'] else None,
        )

    def _load_file(self) -> str:
        """Remove BOM and HOUSE OVERSIGHT lines, strip whitespace."""
        body = self.eml.get_body(('plain', 'related', 'html')).get_content()
        return self.with_header(f"\n{body}")



"""
Politics related files. By default uninteresting.
"""
from epstein_files.documents.config.doc_cfg import DocCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.helpers.string_helper import join_truthy, quote

DIANA_DEGETTE_CAMPAIGN = "Colorado legislator Diana DeGette's campaign"
TRUMP_DISCLOSURES = f"Donald Trump financial disclosures from U.S. Office of Government Ethics"


POLITICS_CFGS = [
    DocCfg(id='030258', author=ALAN_DERSHOWITZ, description=f'{ARTICLE_DRAFT} Mueller probe, almost same as 030248'),
    DocCfg(id='030248', author=ALAN_DERSHOWITZ, description=f'{ARTICLE_DRAFT} Mueller probe, almost same as 030258'),
    DocCfg(id='029165', author=ALAN_DERSHOWITZ, description=f'{ARTICLE_DRAFT} Mueller probe, almost same as 030258'),
    DocCfg(id='029918', author=DIANA_DEGETTE_CAMPAIGN, description=f"bio", date='2012-09-27'),
    DocCfg(id='031184', author=DIANA_DEGETTE_CAMPAIGN, description=f"invitation to fundraiser hosted by {BARBRO_C_EHNBOM}", date='2012-09-27'),
    DocCfg(id='026248', author='Don McGahn', description=f'letter from Trump lawyer to Devin Nunes (R-CA) about FISA courts and Trump'),
    DocCfg(id='027009', author=EHUD_BARAK, description=f"speech to AIPAC", date='2013-03-03'),
    DocCfg(
        id='019233',
        author='Freedom House',
        date='2017-06-02',
        description=f"'Breaking Down Democracy: Goals, Strategies, and Methods of Modern Authoritarians'",
    ),
    DocCfg(id='026856', author='Kevin Rudd', description=f'speech "Xi Jinping, China And The Global Order"', date='2018-06-26'),
    DocCfg(id='026827', author='Scowcroft Group', description=f'report on ISIS', date='2015-11-14'),
    DocCfg(id='024294', author=STACEY_PLASKETT, description=f"campaign flier", date='2016-10-01'),
    DocCfg(
        id='023133',
        author=f"{TERJE_ROD_LARSEN}, Nur Laiq, Fabrice Aidan",
        date='2014-12-09',
        description=f'The Search for Peace in the Arab-Israeli Conflict',
    ),
    DocCfg(id='033468', description=f'{ARTICLE_DRAFT} Rod Rosenstein', date='2018-09-24'),
    DocCfg(id='025849', author=US_ORG, description=quote('Building a Bridge Between FOIA Requesters & Agencies')),
    # CommunicationCfg(id='031670', author="General Mike Flynn's lawyers", recipients=['Sen. Mark Warner & Richard Burr'], description=f"about subpoena"),
    DocCfg(id='031670', description=f"letter from General Mike Flynn's lawyers to senators Mark Warner & Richard Burr about subpoena"),
    DocCfg(
        id='029357',
        date='2015-01-15',
        date_uncertain='a guess',
        description=f"possibly book extract about Israel's challenges entering 2015",
        duplicate_ids=['028887'],
    ),
    DocCfg(id='010617', description=TRUMP_DISCLOSURES, date='2017-01-20', is_interesting=True, attached_to_email_id='033091'),
    DocCfg(id='016699', description=TRUMP_DISCLOSURES, date='2017-01-20', is_interesting=True, attached_to_email_id='033091'),
]



import logging
import re
from collections import Counter
from dataclasses import asdict, dataclass, field
from datetime import datetime
from email import policy
from email.parser import BytesParser
from pathlib import Path
from typing import ClassVar, Self, Sequence, TypeVar

from rich.align import Align
from inflection import underscore
from rich.console import Console, ConsoleOptions, Group, JustifyMethod, RenderResult
from rich.markup import escape
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text
from rich.table import Table

from epstein_files.documents.config.doc_cfg import DEFAULT_TRUNCATE_TO, DUPE_TYPE_STRS, NO_TRUNCATE, DebugDict, EmailCfg, DocCfg, Metadata
from epstein_files.documents.documents.file_info import FileInfo
from epstein_files.documents.documents.search_result import MatchedLine
from epstein_files.documents.emails.constants import DOJ_EMAIL_OCR_REPAIRS, FALLBACK_TIMESTAMP
from epstein_files.documents.emails.email_header import DETECT_EMAIL_REGEX
from epstein_files.output.epstein_highlighter import highlighter, non_epstein_highlighter
from epstein_files.output.highlight_config import HIGHLIGHTED_CONTACTS, get_style_for_category, get_style_for_name
from epstein_files.output.layout_elements.file_display import BasePanel, FileDisplay
from epstein_files.output.html.builder import VERTICAL_MARGIN
from epstein_files.output.rich import (INFO_STYLE, NA_TXT, SYMBOL_STYLE, add_cols_to_table, build_table, console,
     styled_key_value, prefix_with, styled_dict, wrap_in_markup_style)
from epstein_files.output.site.sites import EXTRACTS_BASE_URL
from epstein_files.people.interesting_people import PERSONS_OF_INTEREST, UNINTERESTING_AUTHORS
from epstein_files.people.names import Name
from epstein_files.util.constant.strings import *
from epstein_files.util.constants import CONFIGS_BY_ID
from epstein_files.util.env import args, site_config
from epstein_files.util.helpers.data_helpers import (date_str, patternize, prefix_keys, remove_zero_time,
     uniquify, uniq_sorted, without_falsey)
from epstein_files.util.helpers.link_helper import link_text_obj
from epstein_files.util.helpers.file_helper import coerce_file_path, file_size_to_str
from epstein_files.util.helpers.string_helper import collapse_newlines, join_truthy, quote
from epstein_files.util.logging import DOC_TYPE_STYLES, FILENAME_STYLE, logger

CHECK_LINK_FOR_DETAILS = 'not shown here, check original PDF for details'
CLOSE_PROPERTIES_CHAR = ']'
HOUSE_OVERSIGHT = HOUSE_OVERSIGHT_PREFIX.replace('_', ' ').strip()
MIN_DOCUMENT_ID = 10477

FILENAME_MATCH_STYLES = [
    'dark_green',
    'green',
    'spring_green4',
]

OCR_REPAIRS = {
    re.compile(r'\.corn\b'): '.com',
    re.compile('ln(adequate|dyke)'): r'In\1',
    'Nil Priell': 'Nili Priell',
}

SUMMARY_TABLE_COLS: list[str | dict] = [
    'Count',
    {'name': 'Has Author', 'style': 'honeydew2'},
    {'name': 'No Author', 'style': 'wheat4'},
    {'name': 'Uncertain Author', 'style': 'royal_blue1 dim'},
    {'name': 'Size', 'justify': 'right', 'style': 'dim'},
]

DEBUG_PROPS = [
    'is_interesting',
    'num_lines',
    'people',
    'timestamp',
]

DEBUG_PROPS_TRUTHY_ONLY = [
    AUTHOR,
    'category',
    'is_empty',
]

METADATA_FIELDS = [
    AUTHOR,
    'file_id',
    'filename',
    'num_lines',
    'timestamp',
]


@dataclass
class Document:
    """
    Base class for all Epstein Files documents.

    Attributes:
        file_path (Path): Local path to file
        extracted_author (Name): who created the text in this file, extracted from the text (AKA "not configured")
        extracted_timestamp (datetime, optional): When the file was originally created, extracted from the text
        file_info (FileInfo): Manages things having to do with the underlying file (paths, URLs, etc.)
        lines (list[str]): Number of lines in the file after all the cleanup
        text (str): Contents of the file
    """
    # Class constants
    INCLUDE_DESCRIPTION_IN_SUMMARY_PANEL: ClassVar[bool] = False
    STRIP_WHITESPACE: ClassVar[bool] = True  # Overridden in JsonFile

    file_path: Path
    # Derived at instantiation time
    extracted_author: Name = None
    extracted_timestamp: datetime | None = None
    file_info: FileInfo = field(init=False)
    lines: list[str] = field(default_factory=list)
    text: str = ''

    def __post_init__(self):
        self.file_info = FileInfo(self.file_path)

        if self.file_info.has_file and not self.file_path.exists():
            raise FileNotFoundError(f"File '{self.file_path}' does not exist!")

        self._set_text(text=self.text or self._load_file())
        self._repair()
        self.extracted_author = None if self.author else self._extract_author()
        self.extracted_timestamp = None if self.timestamp else self._extract_timestamp()

    @classmethod
    def from_file_id(cls, file_id: str | int) -> Self:
        """Alternate constructor that finds the file path automatically and builds a `Document`."""
        return cls(coerce_file_path(file_id))

    @property
    def author(self) -> Name:
        return self.config.author if self.config and self.config.author else self.extracted_author

    @property
    def border_style(self) -> str:
        """Should be overloaded in subclasses."""
        return 'white'

    @property
    def category(self) -> str:
        return self.config.category if self.config and self.config.category else self.default_category()

    @property
    def category_style(self) -> str:
        return get_style_for_category(self.category) or ''

    @property
    def config(self) -> DocCfg | None:
        """Get the configured `DocCfg` object for this file id (if any)."""
        return CONFIGS_BY_ID.get(self.file_id)

    @property
    def config_description(self) -> str:
        """Add parentheses to `self.config.description`."""
        return f"{self.config.description}" if self.config and self.config.description else ''

    @property
    def config_description_txt(self) -> Text | None:
        """Add parentheses to `self.config.description`."""
        style = 'bright_white italic' if site_config.email_info_in_subtitle else INFO_STYLE
        return non_epstein_highlighter(Text(self.config_description, style)) if self.config_description else None

    @property
    def config_replace_text_with(self) -> str | None:
        """Configured replacement text."""
        if self.config and self.config.replace_text_with:
            text = join_truthy(self.config.author, self.config.replace_text_with)

            if len(text) < 300 and not text.startswith('photo'):
                return f"(Text of {text} {CHECK_LINK_FOR_DETAILS})"
            else:
                return text

    @property
    def date_str(self) -> str | None:
        return date_str(self.timestamp)

    @property
    def duplicate_file_txt(self) -> Text | None:
        """If the file is a dupe make a nice message to explain what file it's a duplicate of."""
        if self.is_duplicate:
            # TODO: this link is incorrect for DOJ files
            link = link_text_obj(FileInfo.url_for_id(self.file_info.url_slug), self.file_id, 'royal_blue1')
            return self._skipped_file_txt(Text(f"{DUPE_TYPE_STRS[self.config.dupe_type]} ").append(link))

    @property
    def duplicate_of_id(self) -> str | None:
        if self.config and self.config.duplicate_of_id:
            return self.config.duplicate_of_id

    @property
    def empty_file_txt(self) -> Text | None:
        """Overridden in DojFile."""
        pass

    @property
    def external_link_markup(self) -> str:
        """Rich markup string with link to source document."""
        return self.file_info.external_link_markup(get_style_for_name(self.author) if self.author else '')

    @property
    def external_link_txt(self) -> Text:
        return Text.from_markup(self.external_link_markup)

    @property
    def file_id(self) -> str:
        return self.file_info.file_id

    @property
    def file_id_panel(self) -> BasePanel:
        """The header panel printed before the body and subheaders with links and file ID etc."""
        return BasePanel(border_style=self.border_style, text=self.colored_external_links())

    @property
    def filename(self) -> str:
        return self.file_info.filename

    @property
    def html_margin_bottom(self) -> str:
        """Overloaded in `Email` for case of emails with attachments."""
        return VERTICAL_MARGIN

    @property
    def info(self) -> list[Text]:
        """0 to 2 sentences containing the info_txt() as well as any configured description."""
        return without_falsey([self.subheader, self.config_description_txt])

    @property
    def is_attachment(self) -> bool:
        return bool(self.config and self.config.attached_to_email_id)

    @property
    def is_attribution_uncertain(self) -> bool:
        return bool(self.config and self.config.is_attribution_uncertain)

    @property
    def is_duplicate(self) -> bool:
        return bool(self.duplicate_of_id)

    @property
    def is_email(self) -> bool:
        """True if the text looks like it's probably an email."""
        search_area = self.repair_ocr_text(DOJ_EMAIL_OCR_REPAIRS, self.text[0:5000])
        return isinstance(self.config, EmailCfg) or bool(DETECT_EMAIL_REGEX.match(search_area) and self.config is None)

    @property
    def is_empty(self) -> bool:
        return len(self.text.strip()) == 0

    @property
    def is_interesting(self) -> bool | None:
        """
        If `self.config.is_of_interest` returns a bool use that, otherwise house oversight files are
        interesting if they are empty and uninteresting authors are uninteresting.
        Checking of `TEXTERS_OF_INTEREST` etc. is left to the relevant subclass in cases where this
        function returns None.
        """
        if self.config and (is_of_interest := self.config.is_of_interest) is not None:
            return is_of_interest
        elif self.is_attachment:
            return False
        elif self.author in UNINTERESTING_AUTHORS:
            return False
        elif self.file_info.is_house_oversight_file and not (self.author or self.config):
            return True
        elif self.config and self.config.show_with_name in PERSONS_OF_INTEREST:
            return True

    @property
    def length(self) -> int:
        return len(self.text)

    @property
    def metadata(self) -> Metadata:
        metadata = self.config.metadata if self.config else {}
        metadata.update({k: getattr(self, k) for k in METADATA_FIELDS if getattr(self, k) is not None})
        metadata['file_size'] = self.file_info.file_size
        metadata['type'] = self._class_name

        if self.file_info.is_local_extract_file:
            metadata['extracted_file'] = {
                'explanation': 'manually extracted from one of the other files',
                'extracted_from': self.file_info.url_slug + '.txt',
                'url': f"{EXTRACTS_BASE_URL}/{self.filename}",
            }

        return metadata

    @property
    def num_lines(self) -> int:
        return len(self.lines)

    @property
    def panel_title_timestamp(self) -> str:
        """String placed in the `title` of the enclosing `Panel` when printing this document's text."""
        if self.timestamp in [None, FALLBACK_TIMESTAMP]:
            return ''

        if self.config and self.config.timestamp:
            prefix = 'approximate' if self.config.date_uncertain else ''
        else:
            prefix = 'inferred'

        return join_truthy(prefix, f"timestamp: {remove_zero_time(self.timestamp)}")

    @property
    def people(self) -> list[str]:
        """Names of people who either sent/received this email or are mentioned in it."""
        text = self.text

        if self.config:
            if not self.config.is_valid_for_name_scan:
                return []
            elif self.config.replace_text_with:
                text = self.config.replace_text_with

            if self.config_description:
                text = f"{self.config_description}\n{text}"

        people = [c.name for c in HIGHLIGHTED_CONTACTS if c.highlight_regex.search(text)]
        people = uniq_sorted(people + ([self.author] if self.author else []))
        non_particpants = self.config.non_participants if self.config else []
        people = [p for p in people if p not in non_particpants]

        if people:
            self.log(f"people() found: {', '.join(people)}")

        return people

    @property
    def prettified_text(self) -> Text:
        """Returns the string we want to print as the body of the document."""
        style = INFO_STYLE if self.config_replace_text_with and len(self.config_replace_text_with) < 300 else ''
        text = self.config_replace_text_with or self.text

        if self.config and self.config.highlight_quote:
            raise ValueError(f"highlight_quote functionality not implemented in Document yet {self}")

        if args.char_nums:
            text = self._inject_line_numbers(text, args.char_nums)

        if self.config is None or self.config.truncate_at in [None, NO_TRUNCATE] or args.whole_file:
            return highlighter(Text(text, style))

        char_range = list(self.config.truncate_at) if self.config.is_excerpt else [0, self.config.truncate_at]
        trim_footer_txt = self.truncation_note(char_range[1])
        chars = text[char_range[0]:char_range[1]]

        if char_range[0] > 0:
            # Add paragraph separators if very long lines
            if max(len(line) for line in  chars.split('\n')) > 120:
                chars = chars.replace('\n', '\n\n')

            _txt = Text('', style=EXCERPT_STYLE)
            _txt.append(f'<...trimmed first {char_range[0]:,} characters...>\n\n', 'dim')
            txt = _txt.append('...').append(highlighter(Text(chars, style)))
        else:
            txt = highlighter(Text(chars, style))

        return txt.append('...\n\n').append(trim_footer_txt)

    @property
    def suppressed_txt(self) -> Text | None:
        """`Text` object to print if this documents is suppressed for various reasons."""
        return self.uninteresting_txt or self.duplicate_file_txt or self.empty_file_txt

    @property
    def subheader(self) -> Text | None:
        """Secondary info about this file (description, recipients, etc). Overload in subclasses."""
        return None

    @property
    def timestamp(self) -> datetime | None:
        if self.config and self.config.timestamp:
            return self.config.timestamp
        else:
            return self.extracted_timestamp

    @property
    def timestamp_sort_key(self) -> tuple[datetime, str, int]:
        """Sort by timestamp, file_id, then whether or not it's a duplicate file."""
        if self.duplicate_of_id:
            sort_id = self.duplicate_of_id
            dupe_idx = 1
        else:
            sort_id = self.file_id
            dupe_idx = 0

        return (self.timestamp or FALLBACK_TIMESTAMP, sort_id, dupe_idx)

    @property
    def uninteresting_txt(self) -> Text | None:
        """Text to print for uninteresting files."""
        if self.config:
            if self.config.attached_to_email_id:
                return self._skipped_file_txt(f"attached to {self.config.attached_to_email_id}")
            elif args._suppress_uninteresting and self.config.is_interesting is False:
                return self._skipped_file_txt("uninteresting")

    @property
    def _class_name(self) -> str:
        return type(self).__name__

    @property
    def _class_style(self) -> str:
        return DOC_TYPE_STYLES[self._class_name]

    @property
    def _debug_prefix(self) -> str:
        return underscore(self._class_name).lower()

    @property
    def _summary(self) -> Text:
        """Summary of this file for logging. Subclasses should extend with a method that closes the open '['."""
        txt = Text('').append(self._class_name, style=self._class_style)
        txt.append(f" {self.file_id}", style=FILENAME_STYLE)

        if self.timestamp:
            timestamp_str = remove_zero_time(self.timestamp).replace('T', ' ')
            txt.append(' (', style=SYMBOL_STYLE)
            txt.append(f"{timestamp_str}", style=TIMESTAMP_DIM).append(')', style=SYMBOL_STYLE)

        txt.append(' [').append(styled_key_value('size', Text(str(self.length), style='aquamarine1')))
        txt.append(", ").append(styled_key_value('lines', self.num_lines))

        if self.config and self.config.duplicate_of_id:
            txt.append(", ").append(styled_key_value('dupe_of', Text(self.config.duplicate_of_id, style='cyan dim')))

        return txt

    @property
    def _summary_panel(self) -> Panel:
        """Panelized description() with info_txt(). Used in search results not in production HTML."""
        sentences = [self._summary]

        if self.INCLUDE_DESCRIPTION_IN_SUMMARY_PANEL:
            sentences += [Text('', style='italic').append(h) for h in self.info]

        return Panel(Group(*sentences), border_style=self._class_style, expand=False)

    def colored_external_links(self) -> Text:
        return self.file_info.build_external_links(with_alt_links=True)

    def file_display(self, align: JustifyMethod | None = None) -> FileDisplay:
        """Allows for proper right vs. left justify."""
        body = BasePanel(
            border_style=self.border_style,
            text=self.prettified_text,
            title=Text(f"({self.panel_title_timestamp})", style='dim') if self.panel_title_timestamp else None,
        )

        return FileDisplay(
            background_color=self.config.background_color if self.config else '',
            body_panel=body,
            file_info=self.file_id_panel,
            indent=site_config.info_indent,
            justify=align,
            margin_bottom=self.html_margin_bottom,
            subheaders=self.info,
        )

    def lines_matching(self, _pattern: re.Pattern | str) -> list[MatchedLine]:
        """Find lines in this file matching a regex pattern."""
        pattern = patternize(_pattern)
        return [MatchedLine(line, i) for i, line in enumerate(self.lines) if pattern.search(line)]

    def log(self, msg: str, level: int = logging.INFO) -> None:
        """Log a message with with this document's filename as a prefix."""
        logger.log(level, f"{self.file_id} {self._class_name} {msg}")

    def log_top_lines(self, n: int = 10, msg: str = '', level: int = logging.INFO) -> None:
        """Log first 'n' lines of self.text at 'level'. 'msg' can be optionally provided."""
        separator = '\n\n' if '\n' in msg else '. '
        msg = (msg + separator) if msg else ''
        self.log(f"{msg}First {n} lines:\n\n{self.top_lines(n)}\n", level)

    def print(self, whole_file: bool = False) -> None:
        """Print this object for some suppression message."""
        if self.is_attachment:
            self.warn(f"is an attachment and self.print() was calleed")
            return
        elif (suppressed_txt := self.suppressed_txt):
            console.print(Padding(suppressed_txt, site_config.suppressed_file_padding()))
            return

        # TODO: this approach to forcing whole_file sucks
        old_whole_file_arg = args.whole_file
        args.whole_file = args.whole_file or whole_file
        console.print(self)
        args.whole_file = old_whole_file_arg

    def raw_text(self) -> str:
        """Reload the raw data from the underlying file and return it."""
        if self.file_info.is_eml_file:
            with open(self.file_path, 'rb') as fp:
                return BytesParser(policy=policy.default).parse(fp).as_string()
        else:
            return self.file_path.read_text()

    def reload(self) -> Self:
        """Rebuild a new version of this object by loading the source file from disk again."""
        return type(self)(self.file_path)

    def repair_ocr_text(self, repairs: dict[str | re.Pattern, str], text: str) -> str:
        """Apply a dict of repairs (key is pattern or string, value is replacement string) to text."""
        for k, v in repairs.items():
            if isinstance(k, re.Pattern):
                text = k.sub(v, text)
            else:
                text = text.replace(k, v)

        return text

    def rich_header(self) -> Group:
        """Panel + subheaders with filename linking to raw file plus any additional info about the file."""
        padded_info = [Padding(sentence, site_config.info_padding()) for sentence in self.info]
        return Group(*([self.file_id_panel] + padded_info))
        elements = [panel] + padded_info
        return Group(*[Align(e, 'right') for e in elements])

    def top_lines(self, n: int = 10) -> str:
        """First n lines."""
        return '\n'.join(self.lines[0:n])[:DEFAULT_TRUNCATE_TO]

    def to_html(self) -> str:
        # TODO: this does not include the timestamp for OtherFiles!
        return self.file_display().to_html()

    def truncation_note(self, truncate_to: int) -> Text:
        """String with link to source URL that will replace the text after the truncation point."""
        link_markup = self.external_link_markup
        trim_note = f"<...trimmed to {truncate_to:,} characters of {self.length:,}, read the rest at {link_markup}...>"
        return Text.from_markup(wrap_in_markup_style(trim_note, 'dim'))

    def truthy_props(self, prop_names: list[str]) -> DebugDict:
        """Return key/value pairs but only if the value is truthy."""
        return {prop: getattr(self, prop) for prop in prop_names if getattr(self, prop)}

    def warn(self, msg: str) -> None:
        """Print a warning message prefixed by info about this `Document`."""
        self.log(msg, level=logging.WARNING)

    def _debug_dict(self, as_txt: bool = False, with_prefixes: bool = True) -> DebugDict | Text:
        """Merge information about this document from config, file info, etc."""
        config_info = self.config.truthy_props if self.config else {}
        file_info = self.file_info.as_dict

        if config_info.get('id') == file_info.get('file_id'):
            config_info.pop('id')

        if file_info.get('source_url') == file_info.get('external_url', 'blah'):
            file_info.pop('external_url')

        if with_prefixes:
            config_info = prefix_keys(type(self.config).__name__, config_info)
            file_info = prefix_keys(underscore(FileInfo.__name__), file_info)

        props = {**config_info, **file_info, **self._debug_props()}
        return styled_dict(props) if as_txt else props

    def _debug_props(self) -> DebugDict:
        """Collects props of this object only (not the config or locations)."""
        props = {k: getattr(self, k) for k in DEBUG_PROPS}
        props.update(self.truthy_props(DEBUG_PROPS_TRUTHY_ONLY))

        if self.file_info.file_size > 100 * 1024:
            props['file_size_str'] = self.file_info.file_size_str

        return prefix_keys(self._debug_prefix, props)

    def _debug_txt(self) -> Text:
        """Prettified version of `self._debug_dict()` suitable for printing."""
        txt_lines = styled_dict(self._debug_dict(), sep=': ')
        return prefix_with(txt_lines, ' ', pfx_style='grey', indent=2)

    def _extract_author(self) -> Name:
        """Extended in `Email` subclass to pull from  headers."""
        pass

    def _extract_timestamp(self) -> datetime | None:
        """Should be implemented in subclasses."""
        pass

    def _inject_line_numbers(self, text: str, interval: int) -> str:
        """Inject character numbers markers into `text`. For debugging only."""
        idx = interval
        new_text = text[:idx]

        while idx < len(text):
            new_text += f'\n\n ------ {idx} ------ \n\n'
            end_idx = idx + interval
            new_text += text[idx:end_idx]
            idx = end_idx

        return new_text

    def _load_file(self) -> str:
        """Remove BOM and HOUSE OVERSIGHT lines, strip whitespace."""
        return self.raw_text()

    def _numbered_lines(self) -> str:
        """For logging."""
        return '\n'.join([f"[{i}] {quote(line)}" for i, line in enumerate(self.lines)])

    def _repair(self) -> None:
        """Can optionally be overloaded in subclasses to further improve self.text."""
        text = self.repair_ocr_text(OCR_REPAIRS, self.text.lstrip('\ufeff').strip())  # remove BOM

        lines = [
            line.strip() if self.STRIP_WHITESPACE else line for line in text.split('\n')
            if not (line.startswith(HOUSE_OVERSIGHT) or line.startswith(EFTA_PREFIX))
        ]

        self._set_text(text=collapse_newlines('\n'.join(lines)))

    def _set_text(self, lines: list[str] | None = None, text: str | None = None) -> None:
        """Set `self.text` and `self.lines` based on arguments passed."""
        if lines and text:
            raise RuntimeError(f"[{self.filename}] Either 'lines' or 'text' arg must be provided (got both)")
        elif lines is not None:
            self.text = '\n'.join(lines).strip()
        elif text is not None:
            self.text = text.strip()
        else:
            raise RuntimeError(f"[{self.filename}] Either 'lines' or 'text' arg must be provided (neither was)")

        logger.debug(f"_set_text() set self.text to\n---\n{self.text}\n---")
        self.lines = [line.strip() if self.STRIP_WHITESPACE else line for line in self.text.split('\n')]

    def _skipped_file_txt(self, reason: str | Text) -> Text:
        txt = Text(f"Skipping ", f"{INFO_STYLE} dim").append(self.external_link_txt)
        return txt.append(" because it's ").append(reason)

    def _write_clean_text(self, output_path: Path) -> None:
        """Write self.text to 'output_path'. Used only for diffing files."""
        if output_path.exists():
            if str(output_path.name).startswith(HOUSE_OVERSIGHT_PREFIX):
                raise RuntimeError(f"'{output_path}' already exists! Not overwriting.")
            else:
                logger.warning(f"Overwriting '{output_path}'...")

        with open(output_path, 'w') as f:
            f.write(self.text)

        logger.warning(f"Wrote {self.length} chars of cleaned {self.filename} to {output_path}.")

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        """Default `Document` renderer (Email and MessengerLog override this)."""
        yield self.file_display()

    def __str__(self) -> str:
        return self._summary.plain

    @classmethod
    def default_category(cls) -> str:
        return ''

    @classmethod
    def files_summary_table(cls, title: str | Text, first_col_name: str) -> Table:
        """Empty table with appropriate cols for summarizing groups of files."""
        table = build_table(title)
        cols = [{'name': first_col_name, 'min_width': 14}] + SUMMARY_TABLE_COLS
        add_cols_to_table(table, cols, 'right')
        return table

    @classmethod
    def files_summary(cls, files: Sequence[Self], is_author_na: bool = False) -> dict[str, str | Text]:
        """Summary info about a group of files."""
        file_count = len(files)
        author_count = cls.known_author_count(files)

        # NOTE: Order matters!
        return {
            'count': str(file_count),
            'author_count': NA_TXT if is_author_na else str(author_count),
            'no_author_count': NA_TXT if is_author_na else str(file_count - author_count),
            'uncertain_author_count': NA_TXT if is_author_na else str(len([f for f in files if f.is_attribution_uncertain])),
            'bytes': file_size_to_str(sum([f.file_info.file_size for f in files])),
        }

    @classmethod
    def file_summary_row(cls, files: Sequence[Self], author_na: bool = False) -> Sequence[str | Text]:
        """Turn the values in the `cls.files_info()` dict into a list so they can be used as a table row."""
        return [v for v in cls.files_summary(files, author_na).values()]

    @classmethod
    def known_author_count(cls, docs: Sequence[Self]) -> int:
        """Number of elements of `docs` that have an author attribution."""
        return len([doc for doc in docs if doc.author])

    @staticmethod
    def count_by_month(docs: Sequence['DocumentType']) -> Counter[str | None]:
        return Counter([d.timestamp.date().isoformat()[0:7] if d.timestamp else None for d in docs])

    @staticmethod
    def sort_by_id(docs: Sequence['DocumentType']) -> list['DocumentType']:
        return sorted(docs, key=lambda d: d.file_id)

    @staticmethod
    def sort_by_length(docs: Sequence['DocumentType']) -> list['DocumentType']:
        return sorted(docs, key=lambda d: d.file_info.file_size, reverse=True)

    @staticmethod
    def sort_by_timestamp(docs: Sequence['DocumentType']) -> list['DocumentType']:
        return sorted(docs, key=lambda doc: doc.timestamp_sort_key)

    @staticmethod
    def uniquify(documents: Sequence['DocumentType']) -> Sequence['DocumentType']:
        """Uniquify by file_id."""
        id_map = {doc.file_id: doc for doc in documents}
        return [doc for doc in id_map.values()]

    @staticmethod
    def without_dupes(docs: Sequence['DocumentType']) -> list['DocumentType']:
        """Remove any duplicate documents."""
        return [doc for doc in docs if not doc.is_duplicate]


DocumentType = TypeVar('DocumentType', bound=Document)



from epstein_files.documents.config.doc_cfg import CommunicationCfg, DocCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *


TEXT_MSG_CFGS = [
    DocCfg(id='033434', description=f"{SCREENSHOT} iMessage chat labeled 'Edwards'"),
    DocCfg(
        id='EFTA01620764',
        author=MELANIE_WALKER,
        author_uncertain=True,
        description=f'conversation about crypto philanthropy and Bill Gates being drunk all the time',
        is_interesting=True,
        truncate_to=(250, 3500)
    ),
    DocCfg(id='EFTA01618381', author=RENATA_BOLOTOVA, description=f'iMessage screenshots', author_uncertain='sneaky'),
    DocCfg(id='EFTA01622387', author=RENATA_BOLOTOVA, description=f'iMessage screenshots', author_uncertain='sneaky'),
    DocCfg(id='EFTA01618494', author=RENATA_BOLOTOVA, description=f'iMessage screenshots', author_uncertain='sneaky'),
    DocCfg(id='EFTA01618400', author=RENATA_BOLOTOVA, description=f'iMessage screenshots', author_uncertain='sneaky'),
    DocCfg(id='EFTA01611898', description=f"screenshot of recent contacts in an iPhone"),
    CommunicationCfg(
        id='EFTA02731576',
        description=f"making contemporaneous accusations",
        recipients=[LEON_BLACK],
        show_full_panel=True
    ),
    CommunicationCfg(id='EFTA01611042', author=ED_BOYLE, recipients=[None, MARIA_PRUSAKOVA]),
    CommunicationCfg(id='EFTA02731525', author=LEON_BLACK, author_uncertain=True, show_full_panel=True),
    CommunicationCfg(
        id='EFTA01612665',
        description='sender is "stressed about finding girls", Epstein gives advice on how to find them',
        is_interesting=True
    ),
]



from dataclasses import dataclass
from rich.terminal_theme import TerminalTheme

from epstein_files.output.site.sites import SiteType
from epstein_files.util.env import args
from epstein_files.util.helpers.string_helper import indented

PAGE_TITLE = '   ∞ Michel de Cryptadamus ∞   '
FONT_FAMILY = "Menlo,'DejaVu Sans Mono',consolas,'Courier New',monospace"

STYLE_TEMPLATE = """{stylesheet}
body {{
    background-color: {background};
    color: {foreground};
}}"""


@dataclass
class HtmlTemplate:
    body: str

    def __str__(self) -> str:
        return f"""<!DOCTYPE html>
<html>
<head>
    <title>Epstein {self.page_type()}</title>
    <link rel="icon" type="image/x-icon" href="https://media.universeodon.com/accounts/avatars/109/363/179/904/598/380/original/eecdc2393e75e8bf.jpg" />
    <meta charset="UTF-8">
    {self.js_redirect()}

    <style>
        {STYLE_TEMPLATE}
    </style>
</head>

<body>
    {self.body}
</body>
</html>"""

    def js_redirect(self) -> str:
        if args.mobile:
            return ''
        else:
            return """
    <script type="text/javascript">
        if (/Android|webOS|iPhone|iPad|iPod|BlackBerry|IEMobile|Opera Mini/i.test(navigator.userAgent)) {{
            window.location.href = """ + f'"{SiteType.get_mobile_redirect_url(args._site_type)}";' + """
        }}
    </script>"""

    def page_type(self) -> str:
        # TODO: this should be integrated with sites config
        if args.all_emails:
            return 'Emails'
        elif args.all_emails_chrono:
            return 'Chronological Emails'
        else:
            return 'Curated'


CUSTOM_HTML_TEMPLATE = HtmlTemplate(
    """<div class="doc_container page_container">\n        {code}\n    </div>"""
)

RICH_HTML_TEMPLATE = HtmlTemplate(
    f"""<pre style="font-family:{FONT_FAMILY}"><code style="font-family:inherit">{{code}}</code></pre>""",
)


# Swap black for white
HTML_TERMINAL_THEME = TerminalTheme(
    (0, 0, 0),
    (255, 255, 255),
    [
        (0, 0, 0),
        (128, 0, 0),
        (0, 128, 0),
        (128, 128, 0),
        (0, 0, 128),
        (128, 0, 128),
        (0, 128, 128),
        (192, 192, 192),
    ],
    [
        (128, 128, 128),
        (255, 0, 0),
        (0, 255, 0),
        (255, 255, 0),
        (0, 0, 255),
        (255, 0, 255),
        (0, 255, 255),
        (255, 255, 255),
    ],
)



import re
import logging
import warnings
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import ClassVar, Sequence

import datefinder
from dateutil.parser import parse
from rich.console import Group
from rich.markup import escape
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from epstein_files.documents.document import CLOSE_PROPERTIES_CHAR, Document
from epstein_files.documents.config.config_builder import build_cfg_from_text
from epstein_files.documents.config.doc_cfg import DocCfg, Metadata
from epstein_files.documents.documents.file_info import FileInfo
from epstein_files.output.epstein_highlighter import highlighter
from epstein_files.output.highlight_config import QUESTION_MARKS_TXT, styled_category
from epstein_files.output.html.builder import table_to_html
from epstein_files.output.rich import build_table, console
from epstein_files.people.interesting_people import PERSONS_OF_INTEREST
from epstein_files.util.constant.strings import *
from epstein_files.util.constants import *
from epstein_files.util.helpers.data_helpers import days_between, remove_timezone, uniquify
from epstein_files.util.helpers.file_helper import FILENAME_LENGTH
from epstein_files.util.helpers.string_helper import DATE_LENGTH, collapse_whitespace
from epstein_files.util.env import args, site_config
from epstein_files.util.logging import logger

FIRST_FEW_LINES = 'First Few Lines'
MAX_DAYS_SPANNED_TO_BE_VALID = 10
MAX_EXTRACTED_TIMESTAMPS = 100
MIN_TIMESTAMP = datetime(2000, 1, 1)
MIN_PAGES_TO_TRUNCATE_PREVIEW = 10
TRUNCATED_PREVIEW_LEN = 200

LOG_INDENT = '\n         '
TIMESTAMP_LOG_INDENT = f'{LOG_INDENT}    '
VAST_HOUSE = 'vast house'  # Michael Wolff article draft about Epstein indicator
# There's 3 spaces (2 for padding, one for divider) between each col and then 2 on each side + 1 indent
PREVIEW_COL_WIDTH = console.width - FILENAME_LENGTH - DATE_LENGTH - (3 * 2) - 6

SKIP_TIMESTAMP_EXTRACT = [
    PALM_BEACH_TSV,
    PALM_BEACH_PROPERTY_INFO,
]


@dataclass
class OtherFile(Document):
    """
    File that is not an email, an iMessage log, or JSON data.

    Attributes:
        derived_cfg (DocCfg, optional): a DocCfg object derived from contents of the file
        was_timestamp_extracted (bool): True if the timestamp was programmatically extracted (and could be wrong)
    """
    derived_cfg: DocCfg | None = None
    was_timestamp_extracted: bool = False

    # Class vars
    INCLUDE_DESCRIPTION_IN_SUMMARY_PANEL: ClassVar[bool] = True  # Class var for logging output
    MAX_TIMESTAMP: ClassVar[datetime] = datetime(2022, 12, 31) # Overloaded in DojFile

    def __post_init__(self):
        super().__post_init__()

        if not self.config and (cfg := build_cfg_from_text(self)):
            self.derived_cfg = cfg

    @property
    def config(self) -> DocCfg | None:
        return super().config or self.derived_cfg

    @property
    def config_description(self) -> str:
        """Overloads superclass property."""
        if self.config and self.config.complete_description:
            pfx = 'Excerpt of ' if self.config.is_excerpt else ''
            return f"{pfx}{self.config.complete_description}"
        else:
            return ''

    @property
    def category_txt(self) -> Text:
        """Returns '???' for missing category."""
        # TODO: create synthetic DocCfg so we don't have to handle QUESTION_MARKS return here
        return styled_category(self.category) if self.category else QUESTION_MARKS_TXT

    @property
    def is_interesting(self) -> bool | None:
        """Junk emails are not interesting."""
        if (is_interesting := super().is_interesting) is not None:
            return is_interesting
        elif self.author is None and self.file_info.is_house_oversight_file:
            return True
        elif self.author is not None and self.author in PERSONS_OF_INTEREST:
            return True

    @property
    def is_valid_for_table(self) -> bool:
        """Return True if this file is OK to put in a table in the curated chronological views."""
        return not (self.config and (self.config.is_excerpt or self.config.show_full_panel))

    @property
    def metadata(self) -> Metadata:
        metadata = super().metadata
        metadata['is_interesting'] = self.is_interesting

        if self.was_timestamp_extracted:
            metadata['was_timestamp_extracted'] = self.was_timestamp_extracted

        return metadata

    @property
    def preview_text(self) -> str:
        """Text at start of file stripped of newlinesfor display in tables and other cramped settings."""
        text = self.config_replace_text_with if self.config_replace_text_with else self.text
        text = collapse_whitespace(text)[0:site_config.other_files_preview_chars]

        if text.count('Page') > MIN_PAGES_TO_TRUNCATE_PREVIEW:
            text = text[0:TRUNCATED_PREVIEW_LEN]

        return text

    @property
    def preview_text_highlighted(self) -> Text:
        txt = highlighter(escape(self.preview_text))

        if self.config_replace_text_with:
            txt = highlighter(Text(escape(self.preview_text), style='italic grey35'))
        elif self.length > site_config.other_files_preview_chars:
            num_missing_chars = self.length - len(txt)
            txt.append(f"... ({num_missing_chars:,} more characters)", 'dim italic')

        return txt

    @property
    def _summary(self) -> Text:
        """One line summary mostly for logging."""
        return super()._summary.append(CLOSE_PROPERTIES_CHAR)

    def _extract_timestamp(self) -> datetime | None:
        """Return configured timestamp or value extracted by scanning text with datefinder."""
        timestamps: list[datetime] = []

        if self.config and any([s in self.config_description for s in SKIP_TIMESTAMP_EXTRACT]):
            return None

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", module="dateutil")

            try:
                # TODO: datefinder.find_dates() cannot find 08/29/2019 style e.g. in EFTA00005783 :(
                for timestamp in datefinder.find_dates(self.text, strict=False):
                    timestamp = remove_timezone(timestamp)
                    logger.debug(f"Found timestamp: '{timestamp}'")

                    if MIN_TIMESTAMP < timestamp < self.MAX_TIMESTAMP:
                        timestamps.append(timestamp)

                    if len(timestamps) >= MAX_EXTRACTED_TIMESTAMPS:
                        break
            except ValueError as e:
                self.warn(f"Error while iterating through datefinder.find_dates(): {e}")

        if len(timestamps) == 0:
            if not (self.is_duplicate or VAST_HOUSE in self.text):
                self.log_top_lines(15, msg=f"No timestamps found")

            return None

        self.was_timestamp_extracted = True

        if len(timestamps) == 1:
            return timestamps[0]
        else:
            timestamps = sorted(uniquify(timestamps), reverse=True)
            self._log_extracted_timestamps_info(timestamps)
            return timestamps[0]  # Most recent timestamp appearing in text is usually the closest

    def _log_extracted_timestamps_info(self, timestamps: list[datetime]) -> None:
        num_days_spanned = days_between(timestamps[-1], timestamps[0])
        timestamps_log_msg = f"Extracted {len(timestamps)} timestamps spanning {num_days_spanned} days{TIMESTAMP_LOG_INDENT}"
        timestamps_log_msg += TIMESTAMP_LOG_INDENT.join([str(dt) for dt in timestamps])

        if num_days_spanned > MAX_DAYS_SPANNED_TO_BE_VALID and VAST_HOUSE not in self.text:
            self.log_top_lines(15, msg=timestamps_log_msg, level=logging.DEBUG)

    @classmethod
    def files_preview_table(
        cls,
        files: Sequence['OtherFile'],
        title_pfx: str = '',
        title: str | Text | None = '',
        title_justify: str = '',
        **kwargs
    ) -> Table:
        """Build a table of `OtherFile` documents."""
        title = '' if title is None else (title or f'{title_pfx}Other Files Details in Chronological Order')
        title_justify = title_justify or ('left' if title else 'center')
        table = build_table(title, show_lines=True, title_justify=title_justify, **kwargs)
        table.add_column('File', justify='center', width=FILENAME_LENGTH)
        table.add_column('Info', justify='center', width=DATE_LENGTH)
        table.add_column(FIRST_FEW_LINES, justify='left', style='pale_turquoise4', width=PREVIEW_COL_WIDTH)

        for file in files:
            # call superclass method to avoid border_style rainbow
            link_and_info = [FileInfo.build_external_links(file.file_info, file.category_style, id_only=bool(args.mobile))]

            if file.date_str:
                date_txt = Text(file.date_str, style=TIMESTAMP_STYLE)
            else:
                date_txt = Text('(no date)', style=f"{TIMESTAMP_STYLE} dim italic")

            if file.is_duplicate:
                preview_text = file.duplicate_file_txt
                row_style = 'dim'
            else:
                preview_text = file.preview_text_highlighted
                row_style = ''
                link_and_info += file.info

            table.add_row(
                Group(*link_and_info),
                Group(date_txt, file.category_txt, Text(file.file_info.file_size_str, style='wheat4 dim')),
                preview_text,
                style=row_style
            )

        return cls._mobilize_table(table) if args.mobile else table

    @classmethod
    def files_preview_table_html(
        cls,
        files: Sequence['OtherFile'],
        title_pfx: str = '',
        title: str | Text | None = '',
        title_justify: str = '',
        **kwargs
    ) -> str:
        return table_to_html(cls.files_preview_table(files, title_pfx, title, title_justify, **kwargs))

    @classmethod
    def _mobilize_table(cls, _table: Table) -> Table:
        """Make a mobile version of the `files_preview_table()`."""
        table = build_table(_table.title, copy_props_from=_table)
        table.add_column('File', justify='center', width=len('EFTA00424931'))
        table.add_column(FIRST_FEW_LINES, justify='left', style='pale_turquoise4')
        max_col_idx = len(_table.columns) - 1

        for i, row in enumerate(_table.rows):
            table.add_row(
                # Collapse all but last col into one
                Group(
                    *[_table.columns[j]._cells[i] for j in range(1, max_col_idx)],
                    _table.columns[0]._cells[i],
                ),
                _table.columns[max_col_idx]._cells[i],
                style=row.style
            )

        return table

    @classmethod
    def summary_table(cls, files: Sequence['OtherFile'], title_pfx: str = '') -> Table:
        """Table showing file count by category."""
        categories = uniquify([f.category for f in files])
        categories = sorted(categories, key=lambda c: -len([f for f in files if f.category == c]))
        table = cls.files_summary_table(f'{title_pfx}Other Files Summary', 'Category')

        for category in categories:
            category_files = [f for f in files if f.category == category]
            table.add_row(styled_category(category), *cls.file_summary_row(category_files))

        table.columns = table.columns[:-2] + [table.columns[-1]]  # Removee unknown author col
        return table
